/******************************************************************************
 *
 * cachin.c: All the caching functions. 
 *
 * 
 * ***************************************************************************/


#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>           /*mkdir */
#include "global.h"
#include "md5.h"

#include <sys/socket.h>         /* Network */
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#include <errno.h>
#include <glib.h>

#include "proxyServer.h"
#include "caching.h"
#include "cached_objects_hash.h"
#include "parser.h"

#include "main.h"
#ifndef ARM
	#include <curl/curl.h>          /* URL lib */
	#include <curl/types.h>
	#include <curl/easy.h>
#endif

//#define USE_MD5 0

static void MDPrint(void *signature, char *str, const int str_len);
static void MDString(char *string, char *md5);

extern char cache_dir_cfg[512];

/**
 * This function is used to iterate through all of the 
 * key/value pairs in the hashtable by calling g_hash_table_foreach
 *
 */
void foreach_func1(gpointer key, gpointer value, gpointer user_data)
{
    printf("%s(%d) in foreach: %s:%s\n", __FILE__, __LINE__, 
											(char *)key, (char *)value);
}

/*
 * cache_strip_path()
 */
char * cache_strip_path(char *orig_url)
{
	char *tmp;
	char *result;
	int pos;
	
	pos = strlen(cache_dir_cfg);
	tmp = strstr(orig_url, cache_dir_cfg);
	
	result = &tmp[pos];
	
	orig_url = result;

	return orig_url;
}
/*
 * is_file_in_cache
 * returns 0 if the url is in the cache -1 otherwise
 */
int
is_file_in_cache(struct parsed_request *req,
                 struct data_structures *data, int fd)
{
    gpointer value = NULL;
    pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

    /*
     * If the request url is in the hashtable, pass the value to get
     * _from_cache
     */
    printf("isfileincache() %s\n", req->url);

	value = is_url_in_hash(req->url);
	
    if (value == NULL)
        return -1;
    else
        printf("CACHE-HASH LOOKUP IS %s\n", (char *)value);

    /* Get from cache */
    if (get_from_cache(value, fd) < 0) {
        printf("Error:is_file_in_cache: Couldn't get from cache\n");
        return -1;
    }

    return 0;
}

/*
 * put_in_cache()
 * Places the path of the file under the url key
 */
void put_in_cache(struct data_structures *data, char *path, char *url)
{
	cob_put_in_cache(path, url);   
}

/*
 * get_number_in_cache()
 * Returns the number of elements in the hash
 */
unsigned int get_number_in_cache(struct data_structures *data)
{
	return cob_get_number_in_cache();
}

/*
 * get_from_cache()
 * returns 0 if the value is retrieved from the cache successfully
 * -1 otherwise
 */
int get_from_cache(gpointer value, int fd)
{
    /*
     * Open the filename pointed by the value pointer and feed back
     * the data to the socket
     */
    FILE *file;
    char *path;
    int bytes_sent;
    struct stat file_stat;
    int size;
    char *data;

    path = (char *) value;
    printf("\nPath is %s\n \n value is %s\n", path, (char *)value);
    file = fopen(path, "r");

    if (file == NULL) {
        printf("Error:get_from_cache: Couldn't open %s\n", path);
        return -1;
    }

    /* Stat the file for size */
    if (stat(path, &file_stat) != 0) {
        printf("Error::get_from_cache:stat errno %d\n", errno);
        if (errno == EACCES)
            printf("No access\n");
        else if (errno == ENOENT)
            printf("Path doesn't exist\n");
        else if (errno == EBADF)
            printf("filedes is bad\n");
        else if (errno == EFAULT)
            printf("address is bad\n");
        else if (errno == ENAMETOOLONG)
            printf("name too long \n");

        goto error;
    }

    size = file_stat.st_size;

    /* Allocate enough space for the data */
    data = malloc(size);

    if (data == NULL) {
        printf("Error:get_from_cache:Couldn't allocate %d bytes\n", size);
        goto error;
    }

    if (fread(data, 1, size, file) != size) {
        printf("Error:get_from_cache: Error reading file %s\n", path);
        goto error;
    }

    /* Send to socket */
    bytes_sent = send(fd, data, size, 0);

    if (bytes_sent == -1) {
        printf("Error:get_from_cache: Error sending to socket%d\n", fd);
        goto error;
    }

    /* Free data */
    free(data);
    /* Close file */
    fclose(file);

    return 0;

  error:
    free(data);
    fclose(file);
    return -1;
}

/*
 * save_to_file()
 * This method saves the http page to the cache. Currently, we are making
 * a md5 of the path but the idea is to change it into something simple such
 * as a number that is associated through a Hashtable pointing to the file #
 */
int
save_to_file(struct parsed_request *req, struct MemoryStruct *chunk,
             struct data_structures *data)
{
    FILE *file;
    char const_path[512] = "cache/";
	char cmd[512] = "mkdir -p ";
    char *path;
    char *new_path;
    char *new_afterpath;
    int ret;
    //struct stat sb;
    char *afterpath;
    char md5[33];
	char tmp[512];
	gpointer value = NULL;
    pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	
	/* Lock */ 
	pthread_mutex_lock(&mutex);
	/*
	bzero(md5, sizeof(md5));
	bzero(tmp, sizeof(tmp));
	afterpath = NULL;
	new_afterpath = NULL;
	path = NULL;
	*/
	
    path = &(req->path[0]);
    afterpath = &(req->afterpath[0]);

    new_path = remove_http(path);
    new_afterpath = remove_http(afterpath);

    /*
     * Concatenate fixed path named path...
     * If afterpath doesn't exist create it.
     */
#ifdef USE_MD5
    MDString(new_afterpath, &(md5[0]));
    strcat(&(md5[0]), "/");
	afterpath = strcat(&(const_path[0]), &(md5[0]));
#else
	strncpy(&(tmp[0]), new_afterpath, 512);
    strcat(new_afterpath, "/");
	afterpath = strcat(&(const_path[0]), &(tmp[0]));
	//afterpath = remove_tilda(afterpath);
#endif 
	
	
	printf("Number of elements in cache is %d\n",
							get_number_in_cache(data));

    /*
     * If it's a directory create it
     */
    /*
       if ((ret = mkdir (afterpath, S_IREAD | S_IWRITE | S_IEXEC)) < 0) {
       printf ("Error->save_to_file: couldn't create %s error is %d\n",
       afterpath, ret);
       }
     */
	/*
    if ((ret = mkdir(afterpath, 0777)) < 0) {
        printf("Error->save_to_file: couldn't create %s error is %d\n",
               afterpath, errno);
		
    }*/
	strcat(cmd, afterpath);
	system(cmd);					
	

#ifdef USE_MD5
	path = strcat(&(const_path[0]), &(req->filename[0]));
#else
	path = strcat(&(const_path[0]), &(req->filename[0]));
	//path = remove_tilda(path);
#endif
	
    file = fopen(path, "w+");
    if (file) {
        fwrite(chunk->memory, 1, chunk->size, file);
        fflush(file);
        fclose(file);
        put_in_cache(data, path, &(req->path[0])); /* Put in cache hash table */
		pthread_mutex_unlock(&mutex);
        return 0;
    } else {
        printf("Error: Couldn't open file %s\n", path);
		pthread_mutex_unlock(&mutex);
        return -1;
    }
}
/*
 * MDPrint()
 * Converts digest to ASCII
 */
static void MDPrint(void *signature, char *str, const int str_len)
{

    unsigned char *sig_p;
    char *str_p, *max_p;
    unsigned int high, low;

    str_p = str;
    max_p = str + str_len;
    for (sig_p = (unsigned char *) signature;
         sig_p < (unsigned char *) signature + 16; sig_p++) {

        high = *sig_p / 16;
        low = *sig_p % 16;

        /* account for 2 chars */
        if (str_p + 1 >= max_p) {
            break;
        }

        *str_p++ = HEX_STRING[high];
        *str_p++ = HEX_STRING[low];
    }
    if (str_p < max_p) {
        *str_p++ = '\0';
    }


}

/*
 * MDString()
 * Creates a hexadecimal string representation
 */
static void MDString(char *string, char *md5)
{
    MD5_CTX context;
    unsigned char digest[16];
    unsigned int len = strlen(string);

    MD5Init(&context);
    MD5Update(&context, string, len);
    MD5Final(digest, &context);

    MDPrint(&(digest[0]), md5, 33);
}

