/******************************************************************************
 *
 * configuration_reader.c
 *
 * A configuration reader for 7ds
 *
 * The configuration file can be in found in the configuration directory
 * 
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Change this to point to the config file which contains options */
#define CONFIG_FILE "configuration/config.dat"


/*
 * Globals
 */
int proxy_port_cfg = 8080;
int mcast_port_cfg = 9001;
int unicast_port_cfg = 9000;
int web_port_cfg = 9009;
char cache_dir_cfg[512] = "cache/";
char my_ip_address[16] = "128.59.19.105"; 

/*
 * parse_configuration()
 * 
 * Parses configuration options from configuration file 
 * Returns 0 on success, -1 on failure...
 */
int parse_configuration()
{
	FILE *file;
	char *line = NULL;
	size_t len = 0;
	size_t read;
	char *cp;
	int position;
	int number;
	
	if ((file = fopen(CONFIG_FILE, "r")) == NULL) {
		fprintf(stderr, "Couldn't open configuration file. Using Defaults\n");
		return -1;
	}

	while ((read = getline(&line, &len, file)) != -1) {
		

		/* Proxy Server Port */
		if (strncmp(line, "PROXY_PORT=", strlen("PROXY_PORT=")) == 0) {
			position = strlen("PROXY_PORT=");
			cp = &line[position];
			number = atoi(cp);
			if ((number > 0) && (number < 65536)) {
				proxy_port_cfg = number;
			}
			printf("WebProxy listening on %d\n", proxy_port_cfg);
		}

		/* Multicast Port */
		if (strncmp(line, "MCAST_PORT=", strlen("MCAST_PORT=")) == 0) {
			position = strlen("MCAST_PORT=");
			cp = &line[position];
			number = atoi(cp);
			if ((number > 0) && (number < 65536)) {
				mcast_port_cfg = number;
			}
			printf("Multicast port is %d\n", mcast_port_cfg);
		}
		
		/* Unicast port */
		if (strncmp(line, "UNICAST_PORT=", strlen("UNICAST_PORT=")) == 0) {
			position = strlen("UNICAST_PORT=");
			cp = &line[position];
			number = atoi(cp);
			if ((number > 0) && (number < 65536)) {
				unicast_port_cfg = number;
			}		
			printf("Report listener on port %d\n", unicast_port_cfg);
		}

		/* Webserver port */
		if (strncmp(line, "WEB_PORT=", strlen("WEB_PORT=")) == 0) {
			position = strlen("WEB_PORT=");
			cp = &line[position];
			number = atoi(cp);
			if ((number > 0) && (number < 65536)) {
				web_port_cfg = number;
			}
			printf("WebServer listening on port %d\n", web_port_cfg);
		}


		/* cache dir */
		if (strncmp(line, "CACHE_DIR=", strlen("CACHE_DIR=")) == 0) {
			position = strlen("CACHE_DIR=");
			cp = &line[position];
			strncpy(cache_dir_cfg, cp, strlen(cache_dir_cfg));
			
			printf("Cache directory is %s\n", cache_dir_cfg);			
		}


		//Multicast Address
		if (strncmp(line, "MY_IP_ADDRESS=", strlen("MY_IP_ADDRESS=")) == 0) {
			position = strlen("MY_IP_ADDRESS=");
			cp = &line[position];
			printf("my ip address is %s\n", cp);
			strncpy(my_ip_address, cp, strlen(my_ip_address)+1);		
		}

		

	}

	if (line) free(line);

	return 0;
}

/*
int main()
{
	parse_configuration();

	return 0;
}
*/
