/******************************************************************************
 *
 * glimpse_search_engine.c
 * 
 * A search engine based on glimpse.
 *
 * Glimpse is required...
 *
 *
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/

#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/xmlmemory.h>
#include "xmlparser.h"

#define GLIMPSE "glimpse -y -l -F htm "
#define DEBUG(x) printf(x)

/*
 * get_report()
 *
 * Create xml from keyword results.
 * return 0 on success, -1 otherwise
 */
int get_report(char *keyword, char *xml)
{		
	char result[200];
	char *command;
	FILE *pipein_fp;
	
	/*xml stuff*/
	xmlDocPtr doc;
	xmlNodePtr cur;
	xmlNodePtr query_cur;
	xmlNodePtr root;
	xmlChar *mem;
	int size;
	
	/*Concatenate command and keyword*/
	
	command = malloc(sizeof(GLIMPSE) + strlen(keyword));

	if (command == NULL) {
		fprintf(stderr, "get_report:: Not enough mem for command\n");
		return -1;
	}
	
	strncpy(command, GLIMPSE, sizeof(GLIMPSE));
	strncat(command, keyword, strlen(keyword));

	/*
	 * Create XML doc
	 */
	doc = xmlNewDoc((const xmlChar *) "1.0");
	
	if (doc == NULL ) {
		fprintf(stderr,
				"get_report:: Document not created successfully.\n");
		goto cleanup;
	}

	root = xmlNewNode(NULL, (const xmlChar *) XML_ROOT);
	
	xmlDocSetRootElement(doc, root);
	
	if (root == NULL) {
			fprintf(stderr,
						"get_report:: Root node not created\n");
		goto cleanup;

	}

	/*
	 * Add query structure
	 */
	if ((cur = create_xml_query(root, query_cur, keyword)) == NULL) {
		fprintf(stderr, "get_report:: Couldn't creat xml doc\n");
		goto cleanup;
	}


	/*
	 * Add query name to xml report
	 */
	if (create_xml_query_name(cur, query_cur, keyword) < 0) {
		fprintf(stderr, "get_report:: Couldn't creat xml doc\n");
		goto cleanup;
	}
	
   /*
	* Pipe output of glimpse search
	*/
	if ((pipein_fp = popen(command, "r")) == NULL) {
		perror("popen");
		goto cleanup;
	}
	
	/*
	 * Each line is a new report. Parse line for appropriate information.
	 */
	while (fgets(result, sizeof(result), pipein_fp) != NULL) {

		printf("Result: %s\n", result);

		/*
		 * Create Report
		 */	
		if (create_xml_report(cur, query_cur, result) < 0) {
			fprintf(stderr, "get_report:: Couldn't creat xml doc\n");
			pclose(pipein_fp);
			goto cleanup;
		}

		
		/*
		 * Add local url
		 */
		//create_xml_local_url(root, query_cur, result);

		/*
		 * Add more fields for each report as needed below
		 * FIXME: Add more fields
		 */
		
	}	
	
	/* Copy to xml */
	xmlDocDumpMemory(doc, &mem, &size);	
	/*Make sure you free xml after use */
	//xml = malloc(size + 1);
	strncpy(xml, mem, size); 
	printf("Mem contents are \n%s\n", xml);


	pclose(pipein_fp);
	xmlFreeDoc(doc);
	free(command);
	return 0;

cleanup:
	xmlFreeDoc(doc);
	free(command);
	return -1;
}

/*
 *
 *
 *
 * Query>
<QueryName>sports</QueryName>

<Report>
<Title>12 ining game</Title>
<Orig-URL>http://www.sports.com</Orig-URL>
<Date>11 December 2000</Date>
<Summary>NOPe, sorry</Summary>
<readable>1</readable>
</Report>
*/
