
/******************************************************************************
 *
 * guimain.c : 
 *  
 *  This "class" is responsible for parsing the gui commands as the come in
 * through the browser and taking the correct actions...
 * 
 * ***************************************************************************/

#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>           /*mkdir */
#include "global.h"
#include "md5.h"

#include <sys/socket.h>         /* Network */
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#include <errno.h>


#include "proxyServer.h"
#include "guimain.h"
#include "generate_search.h"
#include "multicast.h"
#include "query_report_buffer.h"
#include "parser.h"

#ifndef ARM
	#include <curl/curl.h>          /* URL lib */
	#include <curl/types.h>
	#include <curl/easy.h>
#endif

static void debug(char *msg)
{
	printf("guimain:: %s\n", msg);
}

/*
 * is_query_in_buffer()
 *
 * FIXME: DEPRECATED:
 * Checks if given query is in the query_report_buffer
 * returns 0 if the query is in buffer
 * 		   -1 if the query is NOT in the buffer
 */
int is_query_in_buffer(char *query, struct data_structures *data)
{
	gpointer value = NULL;
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	
	pthread_mutex_lock(&mutex);
	
	/*FIXME Lookup query */	
	
    pthread_mutex_unlock(&mutex);

	if (value == NULL)
		return -1;
	else
		debug("qrb returns\n");

	return 0;
	
}

/*
 * add_query()
 * Add query to query_report_buffer
 * Returns -1 on failure, 0 on success
 */
int add_query(char *desc)
{
	query_t *query;
		
	query = qrb_create_query();
	
	if (query == NULL) {
		fprintf(stderr, "add_query:: Couldn't create query\n");
		return -1;
	}
	
	strncpy(query->descriptor, desc, sizeof(query->descriptor));

	qrb_append_queries(query);

	return 0;
}

/*
 * query_send()
 * Sends the query to the ad hoc network if the query has not
 * already been serviced
 */
void query_send(int fd, struct parsed_request *req, 
					struct data_structures *data)
{
	pthread_t query_multicast_thread;
	
	if ((req->search_query == NULL) || (strlen(req->search_query) == 0)) {
		debug("Query is empty or null\n");
		return;
	}	

	/* If this is a new query */
	if (qrb_is_query_in_buffer(req->search_query) == -1) {

		/* add query */
		add_query(req->search_query);	
		    
		if ((pthread_create(&query_multicast_thread,
                            NULL,
                            (void *(*)(void *)) multicast_keyword,
                            (void *) &req->search_query)) != 0) {
            	printf("Error -> ProxyServer: Error creating thread\n");
        }

		/* Create xml message with query word or url*/
		//multicast_keyword(req->search_query);

		/* Sleep for 1.5 seconds while we wait for reports to come in*/
		sleep(1.5); 
		
		generate_search(fd, data); /*create 7ds page */
	} else {
		debug("Query is NOT new...\n");
		generate_search(fd, data);
		return;
	}
	
}

/*
 * http_query_send()
 * Sends http queries in the form of http://www.columbia.edu/
 */
void http_query_send(int fd, struct parsed_request *req, 
					struct data_structures *data)
{
	pthread_t http_multicast_thread;
	
	if ((req->url == NULL) || (strlen(req->url) == 0)) {
		debug("url is empty or null\n");
		return;
	}	
	if (is_request_url(req->url) == 1) {
		/* If this is a new query */
		if (qrb_is_query_in_buffer(req->url) == -1) {

			/* add query */
			add_query(req->url);	
		    
			if ((pthread_create(&http_multicast_thread,
                            NULL,
                            (void *(*)(void *)) multicast_keyword,
                            (void *) &req->url)) != 0) {
            	printf("Error -> ProxyServer: Error creating thread\n");
        	}

			/* Create xml message with query word or url*/
			//multicast_keyword(req->url);

			/* Sleep for 1.5 seconds while we wait for reports to come in*/
			sleep(1.5); 
			generate_search(fd, data); /*create 7ds page */
		} else {
			debug("Query is NOT new...\n");
			generate_search(fd, data);
			return;
		}
		
	}//is_request
}

/*
 * parse_control()
 * 
 * Parse the control messages such as expand and delete...
 *
 * This is where the messages from the 7ds_control interface are
 * interpreted...
 */
static void parse_control(struct parsed_request *req, int fd)
{
	char *control;
	char *url;
	char *t;
	FILE *picture;
	char buf[1000];

	url = &(req->url[0]);
	control = remove_control(url);

	if (control == NULL)
		return;

	printf("command %s\n", control);

	t = strtok(control, " ");
	
	/* DELETE QUERIES */
	if (strcmp(t, "DELETE") == 0) {
		printf("Deleting Query");
		t = strtok(NULL, " ");
		printf("Number %s\n", t);
		qrb_delete_query(t);
	}

	/* SORT_DATE QUERY */	
	else if (strcmp(t, "SORT_DATE") == 0) {
		t = strtok(NULL, " ");
		qrb_query_set_sort(t, SORT_DATE);
		printf("Sorting query:%s by date\n", t);
	}
	/* SORT_RANK QUERY */
	else if (strcmp(t, "SORT_RANK") == 0) {
		t = strtok(NULL, " ");
		qrb_query_set_sort(t, SORT_RANK);
		printf("Sorting query:%s by rank\n", t);
	}
	/* LOAD MAIN PICTURE */
	else if (strcmp(control, "sevends.jpg") == 0) {
		picture = fopen("sevends.jpg", "r");

		if (picture == NULL) {
			perror("Error loading sevends picture");
			return;
		}

		while (fread(buf, 1, sizeof(buf), picture) > 0) {
			send(fd, buf, sizeof(buf), 0);
		}
	}
	/* Load Star icon */
	else if (strcmp(control, "star.ico") == 0) {
		picture = fopen("star.ico", "r");

		if (picture == NULL) {
			perror("Error loading star picture");
			return;
		}

		while (fread(buf, 1, sizeof(buf), picture) > 0) {
			send(fd, buf, sizeof(buf), 0);
		}
	}

	
	
	return;
}


/*
 * Process a request which is to be processed by the local system.
 * This method does not handle requests which should be sent 
 * to the origin (remote) server.
 */
static int parse_and_decide(int fd, struct parsed_request *req,
               struct data_structures *data)
{
	/*Here we parse the command that are in from 7ds control */

	if (is_request_url(req->url) == 1) {
		/* Only try to parse 7ds control methods */
		if (is_request_query(req->url) == 1) {
			/*Standard http request, just send out*/
			query_send(fd, req, data);
			return 0;	
		} else {
			/* If it's a query process it and quickly return */
			if (is_request_control(req->url) == 1) {
				//query_send(fd, req, data);
				parse_control(req, fd);
				generate_search(fd, data);
				debug("7ds control...Parse and decide\n");
				return 0;
			} else {
				/*
				 * Doesn't fit into any of the above categories
				 */	
				return -1;
			}
		}
	}	
		
	return 0;
}

/*
 * gui_main_parse()
 * This is the main parsing mechanism for all the requests that come
 * in from the browser
 */
void gui_main_parse(int fd, struct parsed_request *req,
               struct data_structures *data)
{
    /*
	 * If parse_and_decide fails then it's a regular 
	 * http query
	 */
	if (parse_and_decide(fd, req, data) == -1) {
 		http_query_send(fd, req, data);	
		debug("parsing failed\n");
	} 

	
}


