/******************************************************************************
 *
 * Main.c
 * This is the main program loop for the new 7ds implementation. We start all 
 * the main loops from this program. Global setting are run from here...
 *  
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/
#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>           /*mkdir */


#include "global.h"
#include "md5.h"

#include <sys/socket.h>         /* Network */
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/time.h>
#include <time.h>
/*Glib libs*/
#include <glib.h>

#include "proxyServer.h"
#include "multicast.h"
#include "unicast_receiver.h"
		#include "xmlparser.h"
#include "cached_objects_hash.h"
#include "query_report_buffer.h"
#include "glimpse_search_engine.h"
#include "sevends_http_server.h"
#include "configuration_reader.h"
#include "caching.h"
#include "swish_manager.h"
#include "parser.h"
#include "referer_hash.h"
#include "main.h"

#ifndef ARM
	#include <curl/curl.h>          /* URL lib */
	#include <curl/types.h>
	#include <curl/easy.h>
	#include "smtp_server.h"
	#include "email_objects_buffer.h"
#endif

/*
 * qrb_init()
 * FIXME:DEPRECATED FUNCTION:
 * Initialize the query_report_buffer data
 */
static void qrb_init(qrb_t *qrb)
{
	qrb->update_freq = 20;
	qrb->broadcast_freq = 100;
	qrb->num_of_broadcasts = 10;
	qrb->buffer_freq = 1;
	qrb->cleancache_period = 2;
	qrb->expires = 24;
	qrb->association_freq = 60;
	qrb->query_time_out = 1000;
	qrb->energy = 0;
	qrb->queries = NULL;
}


void test_timer()
{
	GTimeVal result;

	g_get_current_time(&result);
	
	printf("result:%lu\n", result.tv_sec);

	exit(1);
}

/*
 * main()
 * Main loop for program...
 */
int main(int argc, char *argv[])
{
    /*hash -> url, equal->path */
    struct data_structures data;
	pthread_t 	proxy_thread, 
				multicast_thread, 
				unicast_thread,
			  	swish_manager_thread,
				smtp_server_thread;
	//pthread_t web_thread;
	char read_buf[100];
	
	printf("Welcome to Columbia's 7ds\n");
	printf("--------------------------------\n");	

	parse_configuration();

	printf("--------------------------------\n");	
    /*
       The GLib library can be used in a thread-safe mode by calling
       g_thread_init() before making any other GLib calls. In this
       mode GLib automatically locks all internal data structures as
       needed. This does not mean that two threads can simultaneously
       access, for example, a single hash table, but they can access
       two different hash tables simultaneously. If two different
       threads need to access the same hash table, the application is
       responsible for locking itself.
     */

    if (!g_thread_supported())
        g_thread_init(NULL);

	/* create hash tables */
	cached_objects_hash_init();   
	referer_hash_init();	
	
	create_ip_and_port();
	
	
	//test_timer();
	
	/******************** initialize threads ********************************/

#ifndef ARM
	/*
	 * Proxy server 
	 */
	if ((pthread_create(&proxy_thread,
                            NULL,
                            (void *(*)(void *)) proxy_init,
                            (void *) &data)) != 0) {
            printf("Error -> ProxyServer: Error creating thread\n");
    }

	email_objects_buffer_init();
	//test_libesmtp();
	/*
	 * SMTP server
	 */
	if ((pthread_create(&smtp_server_thread,
                            NULL,
                            (void *(*)(void *)) smtp_server_init,
                            (void *) &data)) != 0) {
	}
#endif

	/*
	 * Multicast Server
	 */
	if ((pthread_create(&multicast_thread,
                            NULL,
                            (void *(*)(void *)) multicast_server_init,
                            (void *) &data)) != 0) {
            printf("Error -> ProxyServer: Error creating thread\n");
	}

	/*
	 * Unicast Server
	 */	
	if ((pthread_create(&unicast_thread,
                            NULL,
                            (void *(*)(void *)) unicast_receiver_init,
                            (void *) &data)) != 0) {
            printf("Error -> ProxyServer: Error creating thread\n");
	}

	/*
	 * Swish Manager
	 */
	if ((pthread_create(&swish_manager_thread,
                            NULL,
                            (void *(*)(void *)) swish_manager_init,
                            (void *) &data)) != 0) {
            printf("Error -> ProxyServer: Error creating thread\n");
	}

	
	
    for (;;) {
        /*do nothing */
		scanf("%s", read_buf);
		if (strncmp(read_buf, "exit", 4) == 0) {
			exit(0);
		} else if (strncmp(read_buf, "log", 3) == 0) {
			//cob_log_data();
			cob_create_from_log();
		} else if (strncmp(read_buf, "email", 5) == 0) {
			//email_show_objects();
			email_upload_objects();
		}
    }
    /* 
     * clean up the hash table, if we had allocated our 
     * keys or values, we would clean them up before 
     * destroying the hash table - g_hash_table_foreach()
     * would be a good candidate for cleaning up keys or
     * values that we had allocated.
     */
    return 0;
}




/**************************OLD CFG******/
/*	
	qrb_init(&qrb);
	
	strncpy(report.title, "cnn", sizeof(report.title));
	strncpy(report.orig_url, "www.cnn.com", sizeof(report.orig_url));
	strncpy(report1.title, "nytimes", sizeof(report1.title));
	strncpy(report1.orig_url, "www.nytimes.com", sizeof(report1.orig_url));

	strncpy(query.descriptor, "election", sizeof(query.descriptor));
	strncpy(query1.descriptor, "NY times", sizeof(query1.descriptor));	
	
	query.reports = g_slist_append(query.reports, &report);
    query1.reports = g_slist_append(query.reports, &report1);

	query.reports = g_slist_append(query.reports, &report);
	query1.reports = g_slist_append(query1.reports, &report1);
	
	qrb.queries = g_slist_append(qrb.queries, &query);
	qrb.queries = g_slist_append(qrb.queries, &query1);
	
	for (iterator = qrb.queries; iterator != NULL; 
								iterator = iterator->next) {
		query_tmp = (query_t *) iterator->data;
		tmp_desc = (char *) query_tmp->descriptor;
		printf("Descriptor : %s\n", tmp_desc);
	}
	*/

	/* Create query report buffer and report structure
	query_report_buffer_init();
	
	strncpy(report.title, "cnn", sizeof(report.title));
	strncpy(report.orig_url, "www.cnn.com", sizeof(report.orig_url));
	
	strncpy(report1.title, "nytimes", sizeof(report1.title));
	strncpy(report1.orig_url, "www.nytimes.com", sizeof(report1.orig_url));

	strncpy(query.descriptor, "election", sizeof(query.descriptor));
	strncpy(query1.descriptor, "NY times", sizeof(query1.descriptor));	

	qrb_append_report(&report);
	qrb_append_report(&report1);

	qrb_append_queries(&query);
	qrb_append_queries(&query1);

	qrb_walk_queries();*/


