/******************************************************************************
 * main.h
 *
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/


/*
 * header file for main
 */

#ifndef _MAIN_H_
#define _MAIN_H_

#include <glib.h>
#include <netinet/in.h>

//#define ARM 1


typedef struct _query_report_buffer {
	int update_freq; 
	int broadcast_freq; 
	int num_of_broadcasts; 
	int buffer_freq; 			/* seconds*/
	int cleancache_period; 			/* seconds*/
	int expires; 				/* hours */
	int association_freq; 
	int query_time_out; 			/* ms*/
	int energy;
	GList *queries;				/*link list of query_t */
}qrb_t;

struct data_structures {
	GHashTable *cached_objects_hash;
    	qrb_t *qrb;				/*query report buffer */	
};

/*
 * Datastructure for reports
 */
typedef struct _report {
	char title[200];
	char orig_url[200];
	char local_url[200];
	time_t age;
	char date[200];
	char summary[200];
	int rank;
	int size;
	int new_query;
}report_t;

/*
 * Holds the data associated with the particular query
 */
typedef struct _query {
	time_t newest_date;
	char newest_url[200];
	char descriptor[200];
	char expanded;
	time_t expires_on;
	int sort_type;
	GList *reports; 
}query_t;

/*
 * Data associated with multicast request
 */
struct multicast_data {
	struct data_structures *data;
	char *msg;
	struct sockaddr_in cliAddr;
};

typedef struct _unicast_data {
	char * mesg;
	struct sockaddr_in client;
}unicast_t;

struct request {
	char * pRequest;
	char * data;
};

struct data {
	int fd; /*socket descriptor */
	struct data_structures *data; 
};

/*
 * Holds information associated with a parsed request coming
 * from the browser
 */
struct parsed_request {
	char	url[256];
	char	filename[256];
	char	hostname[256];
	int		hostport;
	char	referer[256];
	char	type_of_request[20];
	int		content_length;
	char	content_type[256];
	char	authorization[256];
	char 	user_agent[256];
	char	method[20];
	char	path[256];
	char 	afterpath[256];
	char	search_query[256];	
};	

/*
 * Holds the information for the email handling
 */
typedef struct _email_object {
	char	mail_from[256];
	char 	mail_to[256];
	char	body[256];
	GList	*rcpt_to;
	size_t	data_size;
	char	*data;
	char	md5[16];
}email_object_t;

/*
 * Holds From information
 */
typedef struct _parsed_from_field {
	char	from[256];
	size_t	size;
	char	body[256];
}parsed_from_field_t;


enum {SORT_RANK, SORT_DATE};

struct MemoryStruct {
	char *memory; /* chunk of memory */
	size_t size; 
};


void query_init(query_t *query);

#endif /* _MAIN_H_ */
