/******************************************************************************
 *
 * parser.c : This is responsible for parsing the requests that come in from
 * the browser and populating that datastructures.
 * 
 *  
 * 
 * ***************************************************************************/

#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>           /*mkdir */
#include "global.h"
#include "md5.h"

#include <sys/socket.h>         /* Network */
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#include <errno.h>

#include "proxyServer.h"
#include "parser.h"
#include "main.h"
#ifndef ARM
	#include <curl/curl.h>          /* URL lib */
	#include <curl/types.h>
	#include <curl/easy.h>
#endif

#define HTTP "http://"
#define CONTROL "http://7ds_control/"
#define QUERY "http://7ds_control/QuerySearch"
#define INDEX_HTML "index.html"


extern char my_ip_address[16];
extern int web_port_cfg;

char new_path[30];

/*
 * fix_url_ending()
 */
void fix_url_ending(struct parsed_request *req)
{
	char url[256];
	
	if ((!ends_with(req->url, "html")) || (!ends_with(req->url, "htm"))) {
		strncat(url, req->url, strlen(req->url));
		strncat(url, "/", 1);
		strncpy(req->url, url, strlen(url));
	}

}

/*
 * attach_index_to_url()
 * 
 * Attaches .index.html to url request ending with a /
 */
void attach_index_to_url(struct parsed_request *req)
{
	char url[256];
	strncat(url, req->url, strlen(req->url));
	strncat(url, INDEX_HTML, strlen(INDEX_HTML));
	strncpy(req->url, url, strlen(url));
}

/*
 * add_ip_and_port()
 * Adds http://ip:PORT/local_url
 *
 * Returns new path or NULL if unsuccessuful
 */
char * add_ip_and_port(char *local_url, char *complete)
{

	strncat(complete, new_path, strlen(new_path));
	strncat(complete, local_url, strlen(local_url));

	return complete;
}
/*
 * attact_port_to_referer()
 * 
 * Attaches the :9009 to the referer
 * Returns the new pointer
 */
char *attach_port_to_referer(char *referer, char *referer_and_port)
{
	strncat(referer_and_port, referer, strlen(referer));
	strncat(referer_and_port, ":9009", 5);
	
	return referer_and_port;	
}

/*
 * remove_tilda()
 */
char *remove_tilda(char *url)
{
	char *position;
	int index;
	int c = '?';
	
	position = strchr(url, '~');

	if (position == NULL) return url;

	index = url - position;

	if (index < 0) {
		index *= -1;
		url[index] = c;
	}
	return url;
}

/*
 * create_ip_and_port()
 * 
 * Create the path on startup and use it thoughout
 */
void create_ip_and_port() 
{
	char port[7];

	strncpy(new_path, HTTP, strlen(HTTP));
	strncat(new_path, my_ip_address, strlen(my_ip_address));
	strcat(new_path, ":");
	sprintf(port, "%d", web_port_cfg);
	strcat(new_path, port);
	strcat(new_path, "/");

}

/*
 * remove_control()
 * 
 * Removes http://7ds_control
 * 
 * Returns pointer to control message without 7ds_control or msg
 */
char *remove_control(char *control)
{
	char *cp;
	int pos;

	if (strlen(control) == strlen(CONTROL)) return NULL;
	
	if (strncasecmp(control, CONTROL, strlen(CONTROL)) == 0) {
        pos = strlen(CONTROL);
		cp = &control[pos];
        cp += strspn(cp, " \t");
        control = cp;
    }

    return control;
	
}
/*
 * remove_http()
 * Removes http:// from path
 * Returns path minus the "http://"
 */
char *remove_http(char *path)
{
    char *cp;

    if (strncasecmp(path, HTTP, strlen(HTTP)) == 0) {
        cp = &path[7];
        cp += strspn(cp, " \t");
        path = cp;
    }

    return path;
}

/*
 * is_request_url()
 * Returns 1 is request starts with http://, 0 otherwise
 */
int is_request_url(char *request) 
{
	if (strncasecmp(request, HTTP, strlen(HTTP)) == 0) {
		return 1;
	} else {
		return 0;
	}
}

/*
 * is_request_control()
 * Returns 1 is request starts with http://7ds_control, 0 otherwise
 */
int is_request_control(char *request)
{
	if (strncasecmp(request, CONTROL, strlen(CONTROL)) == 0) {
		return 1;
	} else {
		return 0;
	}

}


/*
 * is_request_query()
 * Returns 1 is request starts with http://7ds_control, 0 otherwise
 */
int is_request_query(char *request)
{
	if (strncasecmp(request, QUERY, strlen(QUERY)) == 0) {
		return 1;
	} else {
		return 0;
	}

}




/*
 * ends_with()
 */
inline char ends_with(char *str, char *substr)
{
    return ((strstr(str, substr) + strlen(substr)) == (str + strlen(str)));
}

/*
 * extract_query()
 * If query is in the form of query=keyword, this function
 * returns the keyword. If the '=' is not found it returns a
 * NULL char
 */
char *extract_query(char *query)
{
	char *cp;

	if (strncasecmp(query, "query=", 6) == 0) {
		cp = &query[6];
		return cp;
	} else {
		return NULL;
	}

}

/****************PARSING STUFF*************************************/

/*
 * get_request_line()
 * Reads line by line the request that came in from the socket
 * Returns the line read on success, 
 * 0 on failure
 */
char *get_request_line(char *request, int request_length, int *req)
{
    int i;
    char c;
    int request_idx = (int) *req;

    for (i = request_idx; i < request_length; ++request_idx) {
        c = request[request_idx];
        if (c == '\n' || c == '\r') {
            request[request_idx] = '\0';
            ++request_idx;
            if (c == '\r' && request_idx < request_length &&
                request[request_idx] == '\n') {
                request[request_idx] = '\0';
                ++request_idx;
            }
            *req = request_idx;
            return &(request[i]);
        }

    }
    return (char *) 0;
}

/*
 * get_request_from_socket()
 * This method parses the request and places the extracted fields
 * into the parsed_request datastructure
 * Returns -1 on error, 0 on success
 */
int get_request_from_socket(int fd, struct parsed_request *req)
{
    int c = 0;
    char buf[BUF_SIZE + 1];
    char *line;
    char *method_str;
    char *cp;
    int request_idx = 0;
    /* Initialize the request variables. */
    //char *remoteuser = (char*) 0;
    int method = -1;
    char *path = (char *) 0;
    char *file = (char *) 0;
    char *query = "";
	char *search_query;
    char *protocol = "HTTP/1.0";
    //int status = 0;
    //long bytes = -1;
    //char *req_hostname = (char*) 0;
    char *authorization = (char *) 0;
    char *content_type = (char *) 0;
    int content_length = -1;
    char *cookie = (char *) 0;
    char *host = (char *) 0;
    //time_t if_modified_since = (time_t) -1;
    char *referer = "";
    char *useragent = "";
    char *afterpath = "";

    if ((c = recv(fd, buf, BUF_SIZE, 0) < 0)) {
        printf("Error->get_request_from_socket : Couldn't read socket \n");
        return -1;
    }

	printf("REQUEST\n %s\n", buf);
	
    /* Parse first line */
    method_str = get_request_line(buf, BUF_SIZE, &request_idx);
    path = strpbrk(method_str, " \t\n\r");
	printf("First_path %s\n", path);	
    if (path == (char *) 0) {
        //send_error( 400, "Bad Request", (char*) 0, "Can't parse request." );
        printf("Bad Request Can't parse request.\n");
		return -1;
    }
    *path++ = '\0';
    path += strspn(path, " \t\n\r");
    protocol = strpbrk(path, " \t\n\r");
    if (protocol == (char *) 0) {
        //send_error( 400, "Bad Request", (char*) 0, "Can't parse request." );
        printf("Bad Request Can't parse request.\n");
		return -1;
    }
    *protocol++ = '\0';
    query = strchr(path, '?');
    if (query == (char *) 0)
        query = "";
    else
        *query++ = '\0';

	/* If it's a query extract query information */
	search_query = extract_query(query);	
	if (search_query != NULL) {
			memcpy(req->search_query, 
							search_query, 
							sizeof(req->search_query));
			printf("Search Query %s\n", search_query);
	}
    /* Parse the rest of the request headers. */
    while ((line =
            get_request_line(buf, BUF_SIZE, &request_idx)) != (char *) 0) {
        if (line[0] == '\0')
            break;
        else if (strncasecmp(line, "Authorization:", 14) == 0) {
            cp = &line[14];
            cp += strspn(cp, " \t");
            authorization = cp;
            //printf("authorization is %  \n", authorization);
        } else if (strncasecmp(line, "Content-Length:", 15) == 0) {
            cp = &line[15];
            cp += strspn(cp, " \t");
            content_length = atol(cp);
        } else if (strncasecmp(line, "Content-Type:", 13) == 0) {
            cp = &line[13];
            cp += strspn(cp, " \t");
            content_type = cp;
        } else if (strncasecmp(line, "Cookie:", 7) == 0) {
            cp = &line[7];
            cp += strspn(cp, " \t");
            cookie = cp;
        } else if (strncasecmp(line, "Host:", 5) == 0) {
            cp = &line[5];
            cp += strspn(cp, " \t");
            host = cp;
            memcpy(req->hostname, host, sizeof(req->hostname));
        } else if (strncasecmp(line, "If-Modified-Since:", 18) == 0) {
            cp = &line[18];
            cp += strspn(cp, " \t");
            //FIXME if_modified_since = tdate_parse( cp );
        } else if (strncasecmp(line, "Referer:", 8) == 0) {
            cp = &line[8];
            cp += strspn(cp, " \t");
            referer = cp;
            memcpy(req->referer, referer, sizeof(req->referer));
        } else if (strncasecmp(line, "User-Agent:", 11) == 0) {
            cp = &line[11];
            cp += strspn(cp, " \t");
            useragent = cp;
        }
    }                           /*end while */


    if (strcasecmp(method_str, get_method_str(METHOD_GET)) == 0)
        method = METHOD_GET;
    else if (strcasecmp(method_str, get_method_str(METHOD_HEAD)) == 0)
        method = METHOD_HEAD;
    else if (strcasecmp(method_str, get_method_str(METHOD_POST)) == 0)
        method = METHOD_POST;
    else
        printf("That method is not implemented \n");

    memcpy(req->method, (char *) get_method_str(method),
           sizeof(req->method));

    /* Decode path */
    strdecode(path, path);

    file = strrchr(path, '/');
    *file++;

    afterpath = malloc(strlen(path));
    bzero(afterpath, strlen(path));
	
    memmove(afterpath, path, file - path);
    memcpy(req->afterpath, afterpath, sizeof(req->afterpath));
    memcpy(req->filename, file, sizeof(req->filename));
    memcpy(req->path, path, sizeof(req->path));
    memcpy(req->url, path, sizeof(req->url));

	//printf("path %s\n", path);
	//printf("afterpath %s\n", afterpath);

    return 0;                /*FIXME: make int return */

}

/*
 * get_method_str()
 * This method returns the string representation of the Method Type.
 * We only support GET for now.
 * Returns Method String if m is known; 0 otherwise
 */
char *get_method_str(int m)
{
    switch (m) {
    case METHOD_GET:
        return "GET";
    case METHOD_HEAD:
        return "HEAD";
    case METHOD_POST:
        return "POST";
    default:
        return (char *) 0;
    }
}

/*
 * hexit()
 * Make hexadecimal representation out of char c
 */
int hexit(char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    if (c >= 'a' && c <= 'f')
        return c - 'a' + 10;
    if (c >= 'A' && c <= 'F')
        return c - 'A' + 10;
    return 0;                   /* shouldn't happen, we're guarded by isxdigit() */
}

void de_dotdot(char *file)
{
    char *cp;
    char *cp2;
    int l;

    /* Elide any xxx/../ sequences. */
    while ((cp = strstr(file, "/../")) != (char *) 0) {
        for (cp2 = cp - 1; cp2 >= file && *cp2 != '/'; --cp2)
            continue;
        if (cp2 < file)
            break;
        (void) strcpy(cp2, cp + 3);
    }

    /* Also elide any xxx/.. at the end. */
    while ((l = strlen(file)) > 3
           && strcmp((cp = file + l - 3), "/..") == 0) {
        for (cp2 = cp - 1; cp2 >= file && *cp2 != '/'; --cp2)
            continue;
        if (cp2 < file)
            break;
        *cp2 = '\0';
    }
}



/* 
 * strdecode()
 * Copies and decodes a string.  It's ok for from and to to be the
 * same string.
 */
void strdecode(char *to, char *from)
{
    for (; *from != '\0'; ++to, ++from) {
        if (from[0] == '%' && isxdigit(from[1]) && isxdigit(from[2])) {
            *to = hexit(from[1]) * 16 + hexit(from[2]);
            from += 2;
        } else
            *to = *from;
    }
    *to = '\0';
}


