/******************************************************************************
 *
 * 
 * ProxyServer.c: This is responsible for controlling the requests that 
 * come in thru
 * the webbrowser, spawn new threads and save the data if it's not there...
 *
 * 
 * ***************************************************************************/

#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>           /* mkdir */
#include "global.h"
#include "md5.h"

#include <sys/socket.h>         /* Network */
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#include <errno.h>

#include "proxyServer.h"
#include "caching.h"
#include "guimain.h"
#include "parser.h"
#include "referer_hash.h"
#include "main.h"

#ifndef ARM
	#include <curl/curl.h>          /* URL lib */
	#include <curl/types.h>
	#include <curl/easy.h>
#endif

#define INTERNET_ACCESS 1

/******************************* DECLARATIONS ********************************/
static void off_network_access(int fd, struct parsed_request *req,
                   struct data_structures *data);


static void do_request(void *ptr);



static int make_inet_request(int *ptr, 
			struct parsed_request *req,
			struct data_structures *data);

static void go_ad_hoc(int fd, struct parsed_request *req, 
						struct data_structures *data);
static size_t
write_mem_callback(void *ptr, size_t size, size_t nmemb, void *data);

static void proxy_check_referer(struct parsed_request *req);

extern int proxy_port_cfg; /* from configuration_reader.c */

/******************************* END DECLARATIONS ****************************/


/*
 * do_i_have_internet()
 *
 * Returns 1 if we have access to a base station, 0 for ad hoc
 */
static int do_i_have_internet() 
{	
	return INTERNET_ACCESS;
}

/*
 * do_request()
 * This method examines the request that comes in from the socket and 
 * accordingly acts.
 * @param ptr Void ptr to the socket ptr
 * @return void
 */
static void do_request(void *ptr)
{
    struct data *req_data;
    struct data_structures *data;
    struct parsed_request req;
    int fd;

    req_data = (struct data *) ptr;
    data = (struct data_structures *) req_data->data;
    fd = (int) req_data->fd;

    /*
     * Here's where I do all the request serve work
     */

    /* If it's a 7dscontrol command, process here... */

    /* Parse the proxy request */
    get_request_from_socket(fd, &req);

    /*
     * If it's not addressed to my host
     */

    /*
     * If file is in cache retrieve and send
     */
    if (is_file_in_cache(&req, data, fd) == 0) {
        printf("GOT %s FROM CACHE\n", req.url);
        close(fd);
        return;
    }

    /*
     * Try to route request to Internet
     */
    if (do_i_have_internet() == 0) {
		//go_ad_hoc(fd, &req, data);
		off_network_access(fd, &req, data);
	} else	if (make_inet_request(&fd, &req, data) != 0) {
        //go_ad_hoc(fd, &req, data);
		off_network_access(fd, &req, data);	
	}

    /*
     * Send adhoc 
     */

    /*
     * End connection
     */
    close(fd);
    return;
}

/*
 * go_ad_hoc()
 * This method checks if the query passed on from the browser is a new query
 * and processes the cache-buffer and query-hashtable. Finally, if the query
 * is new off_network_access() is called to broadcast the query over the adhoc
 * network
 */
static void
go_ad_hoc(int fd, struct parsed_request *req, struct data_structures *data)
{
    /* If query is new, process otherwise return. */

    /* check query-hashtable & cache buffer */


    /* place in query-hashtable */

	query_send(fd, req, data);
}


/* 
 * off_network_access()
 * This method is called to access the ad hoc network.
 */
static void
off_network_access(int fd, struct parsed_request *req,
                   struct data_structures *data)
{
    gui_main_parse(fd, req, data);
}



/*
 * make_inet_request()
 * 
 */
static int
make_inet_request(int *ptr,
                  struct parsed_request *req, struct data_structures *data)
{
    int fd = (int) *ptr;
    CURL *curl;
    CURLcode res;
    struct MemoryStruct chunk;
    long header;
	char is_true;
	int pos;

    /* Init memory chunk where the curl client reads data */
    chunk.memory = NULL;
    chunk.size = 0;


    /*
     * FIXME : Check in hashtable to lookup a referers address.
     * For example while you've retrieved a web page from a data
     * holder and are browsing the site you just retrieved, if you
     * follow a link from the site it's good practise to try to get 
     * the page from the web server of the data holder directly
     */
	proxy_check_referer(req);

	/*
	is_true = ends_with(req->url, "/");
	if (is_true)
		printf("Ends with a /\n");*/
/*
	pos = strlen(req->url);
//	if (req->url[pos] == '/')
//		printf("Ends with a /\n");
	is_true = req->url[pos -1];
	if (is_true == '/')
		printf("Ends with a /\n");
	else
		printf("Ends with %s", &is_true);	
*/		

    /*
     * GET REQUEST
     */
    curl = curl_easy_init();
    if (curl) {
        /* Set website */
		/*FIXME: Check to see if index.html */

        printf("Getting %s\n", req->url);
        curl_easy_setopt(curl, CURLOPT_URL, req->url);
        /* set mem callback location */
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_mem_callback);
        /* pass chunk to callback function */
        curl_easy_setopt(curl, CURLOPT_FILE, (void *) &chunk);
        /* get it */
        res = curl_easy_perform(curl);

		/*
		 * Deal with errors here
		 */
        if (res != CURLE_OK) {
            fprintf(stderr, "Error: Curl %d\n", res);
			if (res == CURLE_COULDNT_RESOLVE_HOST) {
				fprintf(stderr, "Error: Curl couldn't resolve host %s\n", req->url);
			} else if (res == CURLE_URL_MALFORMAT) {
				fprintf(stderr, "Error: Curl malformed url %s\n", req->url);
			} else if (res == CURLE_COULDNT_CONNECT) {
				fprintf(stderr, "Error: Curl could NOT connect to %s\n", req->url);
			}

            goto curl_error;
        }

		
        /* parse headers */
        res = curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, header);

        if (header > CURLE_OK) {
            /* there was an error, return and don't save */
            /* For error codes look in -> usr/include/curl/curl.h */
            fprintf(stderr,
                    "Error:make_inet_request Header's not 200 OK.\n");
            goto curl_error;
        }
        /* cleanup */
        curl_easy_cleanup(curl);

    }

    /* Send data back to the Browser */
    send(fd, chunk.memory, chunk.size, 0);

    /* If return is 200 OK. If we're here it's always correct */
    save_to_file(req, &chunk, data);

    /* Free data block */
    free(chunk.memory);

    return 0;                   /* success */

  curl_error:
    curl_easy_cleanup(curl);
    free(chunk.memory);
    return -1;
}



/*
 * check_in_cache()
 *
 * Checks in the cache to see if the element is there. If present, extract
 * from cache.
 *
 * Returns 0 on success, -1 on failure
 */
int check_in_cache(struct parsed_request *req, int fd)
{
    /*Cache buffer */

    /*
     * If the page is in the cache and it has the expired field set
     * try to get from real web server
     */
    return 0;
}


/*
 * proxy_init()
 * This is how I'll call the proxy server. I'll need to pass the hashtables and
 * other functions in here...
 * 
 */
int proxy_init(struct data_structures *data)
{
    pthread_t request_thread;
    int listenfd, connfd;
    socklen_t clilen;
    struct sockaddr_in client_addr, server_addr;
    struct data req_data;
    //int SERVER_PORT = 8080;
    int LISTEN_QUEUE = 25;

    printf("Proxy_init()\n");

    /* create server socket */
    if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        printf("Error opening socket\n");
        return -1;
    }

    /* clear out struct */
    bzero(&server_addr, sizeof(server_addr));

    /* populate server address */
    server_addr.sin_family = AF_INET;
    server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    server_addr.sin_port = htons(proxy_port_cfg);

    if (bind
        (listenfd, (struct sockaddr *) &server_addr, sizeof(server_addr))
        < 0) {
        printf("Proxy_init: Error binding socket\n");
        return -1;
    }

    if (listen(listenfd, LISTEN_QUEUE) < 0) {
        printf("Error listening socket\n");
        return -1;
    }

    for (;;) {
        clilen = sizeof(client_addr);
        connfd =
            accept(listenfd, (struct sockaddr *) &client_addr, &clilen);
        req_data.fd = connfd;
        req_data.data = data;
        if ((pthread_create(&request_thread,
                            NULL,
                            (void *(*)(void *)) do_request,
                            (void *) &req_data)) != 0) {
            printf("Error -> ProxyServer: Error creating thread\n");
        }
    }

    return 0;

}

/*
 * write_mem_callback ()
 * This method is called by the curl writeback method. It copies into mem
 * the data that is read by the socket. N.B free(mem-memory) when done.
 * returns the number of bytes written to memory. 
 */
static size_t
write_mem_callback(void *ptr, size_t size, size_t nmemb, void *data)
{
    register int realsize = size * nmemb;
    struct MemoryStruct *mem = (struct MemoryStruct *) data;

    mem->memory = (char *) realloc(mem->memory, mem->size + realsize + 1);

    if (mem->memory) {
        memcpy(&(mem->memory[mem->size]), ptr, realsize);
        mem->size += realsize;
        mem->memory[mem->size] = 0;
    }
    return realsize;
}

/*
 * proxy_check_referer()
 *
 * Checks to see if there's a referer that is in the 7ds 
 * network. If it is then attact the referer to the url
 * request. we create
 * a hashtable of hosts that have replied and check against that. 
 * If we go with the later, we'll have to set timeouts.
 * 
 * There are a few special cases that we examine here. 
 * If the request contains
 * a tilda then we examine the request from the root of the website. 
 * Also, if the 
 * website already calls the full name of the of the link, detect this 
 * and avoid fixing
 * the relative path.
 * 
 */
static void proxy_check_referer(struct parsed_request *req)
{
	char *cp;
	char *data, *path;
	char is_true;
	char *new_path;
	int pos;
	char new_url[256];


	/*
	 * If referer is not in the hashtable or has timed out don't
	 * process request
	 */	 
	if (referer_is_key_in_hash(req->referer) == 0) {
		printf("proxy_check_referer: %s Not a referer\n", req->referer);
		return;
	}
	
	/* Data */	
	cp = strstr(req->url, req->hostname);

	if (cp != NULL) {
		pos = strlen(req->hostname);
		data = &cp[pos];
		printf("Data: %s\n", data);
	}

	/* Path */
	printf("hostname:%s\n", req->hostname);
	cp = NULL;
	cp = strstr(req->referer, req->hostname);
	printf("cp:%s\n", cp);
	if (cp != NULL) {
		pos = strlen(req->hostname);
		path = &cp[pos];
		printf("Path: %s\n", path);
	} else {
		/* Page is not directly related */
		bzero(new_url, sizeof(new_url));
		strcat(new_url, "http://");
		cp = strstr(req->referer, "http://");
		//strcat(new_url, cp);
		cp = strtok(cp + 7, "/");
		strcat(new_url, cp);
//		strcat(new_url, req->referer);
		strcat(new_url, "/");
		strcat(new_url, req->hostname);
		strcat(new_url, data);
		strcpy(req->url, new_url);	
		return;	
	}

	cp = NULL;
	is_true = ends_with(path, "index.html");
	if (is_true) {
		printf("Ends with index.html\n");
		cp = strstr(path, "index.html");
		pos = path - cp;
		new_path = strdup(path);
		if (pos < 0) {
			pos *= -1;
			new_path[pos - 1] = '\0';
		}
		printf("Path now is:%s\n", new_path);
	
		if (strncmp(path, data, strlen(new_path)) == 0) {
			printf("same host\n");
			return;
		}

	} else {
		new_path = strdup(path);		
		pos = strlen(new_path);
		new_path[pos -1] = '\0';
	}
	
	if (strncmp(path, data, strlen(path)) == 0) {
		printf("same host\n");
		return;
	}

	if (strncmp(data, "/~", 2) == 0) {
		printf("tilda\n");
		cp = index(&path[1], '/');
		pos = (path ) - cp;
		if (pos < 0) {
			pos *= -1;
			new_path[pos] = '\0';
		}
	}

	
	bzero(new_url, sizeof(new_url));
	strcat(new_url, "http://");
	strcat(new_url, req->hostname);
	strcat(new_url, new_path);
	strcat(new_url, data);

	strcpy(req->url, new_url);
	printf("------------> New URL:%s\n", new_url);

	return;
}

/*
 * test_referer()
 *
 * This is just a test function for adding referrers
 */
void test_referer()
{
	char *referer = "http://128.59.19.105:9009/www.cs.columbia.edu/";
	char *host = "128.59.19.105:9009";
	char *url = "http://128.59.19.105:9009/~hgs/images/nav.gif";
	char *cp;
	char *data, *path;
	char is_true;
	char *new_path;
	int pos;
	char new_url[256];

	/* Data */	
	cp = strstr(url, host);

	if (cp != NULL) {
		pos = strlen(host);
		data = &cp[pos];
		printf("Data: %s\n", data);
	}

	cp = NULL;

	cp = strstr(referer, host);

	if (cp != NULL) {
		pos = strlen(host);
		path = &cp[pos];
		printf("Path: %s\n", path);
	}

	cp = NULL;
	is_true = ends_with(path, "index.html");
	if (is_true) {
		printf("Ends with index.html\n");
		cp = strstr(path, "index.html");
		pos = path - cp;
		new_path = strdup(path);
		if (pos < 0) {
			pos *= -1;
			new_path[pos - 1] = '\0';
		}
		printf("Path now is:%s\n", new_path);
	} else {
		new_path = strdup(path);		
		pos = strlen(new_path);
		new_path[pos -1] = '\0';
	}

	if (strncmp(path, data, strlen(path)) == 0) {
		printf("same host\n");
		return;
	}

	if (strncmp(data, "/~", 2) == 0) {
		printf("tilda\n");
		cp = index(&path[1], '/');
		pos = (path ) - cp;
		if (pos < 0) {
			pos *= -1;
			new_path[pos] = '\0';
		}
	}
	
	bzero(new_url, sizeof(new_url));
	strcat(new_url, "http://");
	strcat(new_url, host);
	strcat(new_url, new_path);
	strcat(new_url, data);

	printf("New URL:%s\n", new_url);

}
