/******************************************************************************
 * query_report_buffer.c
 *  
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/

/*Glib libs*/
#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>

#include "global.h"
#include "md5.h"
#include "main.h"
#include "query_report_buffer.h"



GList *query_report_buffer;
GList *reports;

/*
 * QRB variables...
 * FIXME: insert methods for changing all this
 */
int qrb_update_freq = 20; 
unsigned int qrb_broadcast_freq = 10; 
int qrb_num_of_broadcasts = 10; 
int qrb_buffer_freq =1; 				/* seconds*/
int qrb_cleancache_period = 2; 			/* seconds*/
int qrb_expires = 24; 					/* hours */
int qrb_association_freq = 60; 
int qrb_query_time_out = 1000; 			/* ms*/
int qrb_energy = 0;

/*
 * get_qrb_update_freq()
 */
int get_qrb_update_freq(void)
{
    return qrb_update_freq;
}

/*
 * get_qrb_broadcast_freq()
 */
unsigned int get_qrb_broadcast_freq(void)
{
	return qrb_broadcast_freq;
}

/*
 * get_qrb_num_of_broadcasts()
 */
int get_qrb_num_of_broadcasts(void)
{
	return qrb_num_of_broadcasts;
}

/*
 * get_qrb_buffer_freq()
 */
int get_qrb_buffer_freq(void)
{
	return qrb_buffer_freq;
}

/*
 * get_qrb_cleancache_period()
 */
int get_qrb_cleancache_period(void)
{
	return qrb_cleancache_period;
}

/*
 * get_qrb_expires()
 */
int get_qrb_expires(void)
{
	return qrb_expires;
}

/*
 * get_qrb_association_freq()
 */
int get_qrb_association_freq(void)
{
	return qrb_association_freq;
}

/*
 * get_qrb_query_time_out()
 */
int get_qrb_query_time_out(void)
{
	return qrb_query_time_out;
}

/*
 * get_qrb_energy()
 */
int get_qrb_energy(void)
{	
	return qrb_energy;
}

/*
 * query_report_buffer_init()
 * Initialize list
 */
void query_report_buffer_init(void)
{
	query_report_buffer = NULL;
	reports = NULL;
}


/*
 * qrb_get_buffer()
 * get pointer to query report buffer
 */
GList* qrb_get_buffer(void)
{
	return query_report_buffer;
}

/*
 * qrb_get_reports()
 * returns pointer to reports list
 */
GList* qrb_get_reports(void)
{
	return reports;
}

/*
 * query_init()
 * Initialize queries
 */
void query_init(query_t *query)
{
	query->newest_date = (time_t) -1;
	query->expanded = 0;
	bzero(query->newest_url, sizeof(query->newest_url));
	bzero(query->descriptor, sizeof(query->descriptor));
	query->sort_type = SORT_RANK;
	query->reports = NULL;
}

void report_init(report_t *report)
{
	bzero(report->title, sizeof(report->title));
	bzero(report->orig_url, sizeof(report->orig_url));
	bzero(report->local_url, sizeof(report->local_url));
	bzero(report->date, sizeof(report->date));
	bzero(report->summary, sizeof(report->summary));
	report->age = (time_t) -1;
	report->new_query = 1;
}

/*
 * qrb_append_report()
 * Append reports to list
 */
void qrb_append_report(report_t *report)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);

	reports = g_list_append(reports, report);

	pthread_mutex_unlock(&mutex);
}

/*
 * qrb_append_queries()
 * Append queries to qrb
 */
void qrb_append_queries(query_t *query)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);

	query_report_buffer = g_list_append(query_report_buffer, query);

	pthread_mutex_unlock(&mutex);

}

/*
 * qrb_is_query_expanded()
 *
 * Returns 1 if there are reports for the query, 0 otherwise
 */
int qrb_is_query_expanded(query_t *query) 
{
	if (query->reports == NULL) {
		return 0;
	} else {
		return 1;
	} 
}

/*
 * qrb_walk_queries()
 * Print all queries
 */
void qrb_walk_queries(void)
{
	GList *iterator;
	query_t *query_tmp;
	char *tmp_desc;	
	
	for (iterator = query_report_buffer; iterator != NULL; 
								iterator = iterator->next) {
		query_tmp = (query_t *) iterator->data;
		tmp_desc = (char *) query_tmp->descriptor;
		printf("Descriptor : %s\n", query_tmp->descriptor);
	}
}

/*
 * delete_this_query()
 * 
 * Deletes the query referenced
 */
static void delete_this_query(query_t *query)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);

	query_report_buffer = g_list_remove(query_report_buffer, query);

	pthread_mutex_unlock(&mutex);

}

void qrb_query_set_sort(char *descriptor, int sort_type)
{
	GList *iterator;
	query_t *query_tmp;
	char *tmp_desc;	
	
	for (iterator = query_report_buffer; iterator != NULL; 
								iterator = iterator->next) {
		query_tmp = (query_t *) iterator->data;
		tmp_desc = (char *) query_tmp->descriptor;
		printf("Sort Descriptor : %s\n", query_tmp->descriptor);
		if (strcmp(query_tmp->descriptor, descriptor) == 0) {
			query_tmp->sort_type = sort_type;
		}
	}

}
/*
 * qrb_delete_query()
 * 
 * Deletes the query
 */
void qrb_delete_query(char *descriptor)
{
	GList *iterator;
	query_t *query_tmp;
	char *tmp_desc;	
	
	for (iterator = query_report_buffer; iterator != NULL; 
								iterator = iterator->next) {
		query_tmp = (query_t *) iterator->data;
		tmp_desc = (char *) query_tmp->descriptor;
		printf("Descriptor : %s\n", query_tmp->descriptor);
		if (strcmp(query_tmp->descriptor, descriptor) == 0) {
			delete_this_query(query_tmp);
		}
	}

}
/*
 * qrb_walk_reports()
 *	Walks through all reports for the query passed.
 */
void qrb_walk_reports(query_t *query)
{
	GList *iterator;
	report_t *report_tmp;
	char *tmp_desc;	

	for (iterator = query->reports; iterator != NULL; 
								iterator = iterator->next) {
		report_tmp = (report_t *) iterator->data;
		tmp_desc = (char *) report_tmp->local_url;
		printf("local url : %s\n", report_tmp->local_url);
	}


}

/*
 * qrb_calcutate_rank()
 */
char *qrb_calculate_rank(int rank) 
{
	if ((rank < 0) || (rank > 1000)) 
		return " ";

	if (0 < rank && rank <= 199) 		return ONE_STAR;
	if (200 <= rank && rank <= 399) 	return TWO_STAR;
	if (400 <= rank && rank <= 599) 	return THREE_STAR;
	if (600 <= rank && rank <= 799) 	return FOUR_STAR;
	if (800 <= rank && rank <= 1000) 	return FIVE_STAR;

	return " ";
}

/*
 * qrb_compare_date()
 *
 * Compares the age of the reports and reports them in Descending
 * order starting from newest to oldest.
 */
gint qrb_compare_date_func(report_t *a, report_t *b)
{
	return b->age - a->age;
}

/*
 * qrb_compare_rank_func()
 *
 * Compares the rank of the web pages and reports them from highest to
 * lowest
 */
gint qrb_compare_rank_func(report_t *a, report_t *b)
{
	return b->rank - a->rank;
}


/*
 * qrb_show_reports()
 *
 * Walks through all the reports of a query and creates a reference
 * that can be accessed through 7ds control
 * Print reports for each query
 */
void qrb_show_reports(char *outbuf, query_t *query)
{
	GList *iterator, *sorted;
	report_t *report_tmp;
	char tmpbuf[1000];

	sorted = g_list_copy(query->reports);
	switch(query->sort_type) {
		case SORT_RANK:
			sorted = g_list_sort(sorted, (GCompareFunc)qrb_compare_rank_func);
			break;
		case SORT_DATE:
			sorted = g_list_sort(sorted, (GCompareFunc)qrb_compare_date_func);
			break;
		default:
			sorted = g_list_sort(sorted, (GCompareFunc)qrb_compare_rank_func);
	}

	
	for (iterator = sorted; iterator != NULL; 
								iterator = iterator->next) {
		report_tmp = (report_t *) iterator->data;
	
		if (report_tmp->new_query == 0) continue;

		bzero(tmpbuf, sizeof(tmpbuf));

		snprintf(tmpbuf, sizeof(tmpbuf),
					"<a href=\"%s\" target=\"%s\" >%s</a> <BR> " 
					"<font color=#00800> %s <BR>-%s-</font> " 
					"%s <BR> <BR>",
					report_tmp->local_url, 
					report_tmp->title, 
					report_tmp->title, 
					report_tmp->local_url,
					report_tmp->date,
					qrb_calculate_rank(report_tmp->rank));

		strcat(outbuf, tmpbuf);	
	}
}

/*
 * qrb_is_report_in_query()
 *
 * Checks if report is already in query. Specifically, it
 * checks if the local url for the same query has been inserted already
 *
 * Returns 1 if query is in report, 0 otherwise
 */
int qrb_is_report_in_query(query_t *query, report_t *report)
{
	GList *iterator;
	report_t *report_tmp;

	for (iterator = query->reports; iterator != NULL; 
								iterator = iterator->next) {
		
		report_tmp = (report_t *) iterator->data;
		if (strcmp(report_tmp->local_url, report->local_url) == 0) {								return 1;
		}
	}
	return 0;
}

/*
 * qrb_remove_duplicate_reports()
 */
void qrb_remove_duplicate_reports(query_t *query, report_t *report)
{
	GList *iterator;
	report_t *report_tmp;
	int new_query=0;
	
	for (iterator = query->reports; iterator != NULL; 
								iterator = iterator->next) {
		
		report_tmp = (report_t *) iterator->data;
		if ((strcmp(report_tmp->title, report->title) == 0) && 
					   (report_tmp->size == report->size))	{		
			new_query = report->age - report_tmp->age;
			if (new_query) {
				report->new_query = 1, report_tmp->new_query = 0;
			}
		}
	}

}
/*
 * qrb_show_queries()
 *
 * Print queries to the GUI. If queries have reports display those as well.
 */
void qrb_show_queries(char *outbuf)
{
	GList *iterator;
	query_t *query_tmp;
	char *tmp_desc;
	char tmpbuf[200];
	int i = 0;
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);

	for (iterator = query_report_buffer; iterator != NULL; 
								iterator = iterator->next) {
			
		query_tmp = (query_t *) iterator->data;
		tmp_desc = (char *) query_tmp->descriptor;
	
		if (query_tmp->reports != NULL) {
			bzero(tmpbuf, sizeof(tmpbuf));
			snprintf(tmpbuf, sizeof(tmpbuf),
					"<TR><TD><FORM><INPUT TYPE=\"button\" VALUE=\"Delete\" " 
					"onclick=\"parent.location='DELETE %s'\"></FORM></TD>\n", 
					query_tmp->descriptor);	

			strcat(outbuf, tmpbuf);
			bzero(tmpbuf, sizeof(tmpbuf));
			snprintf(tmpbuf, sizeof(tmpbuf),
					"<TD align = \"left\"><a href=\"http://7ds_control" 
					"/expand %d\">\n", i);

			strcat(outbuf, tmpbuf);
		
			bzero(tmpbuf, sizeof(tmpbuf));
			snprintf(tmpbuf, sizeof(tmpbuf),
					"<FONT color=\"blue\" size=4><B>Query %d: %s </B></FONT></a>"
					" - <a href=\"http://7ds_control/SORT_DATE %s\"> sort by date</a>" 
					"- <a href=\"http://7ds_control/SORT_RANK %s\"> sort by rank</a>\n",
						i, tmp_desc, tmp_desc, tmp_desc); //FIXME: Add number of reports

			strcat(outbuf, tmpbuf);
					
				if (qrb_is_query_expanded(query_tmp) == 1) {

					/*
				 	 * FIXME: add expanded
			 		 */
					strcat(outbuf, "<TR><TD></TD><TD>");
					qrb_show_reports(outbuf, query_tmp);
					strcat(outbuf, "</TD></TR>\n");
				} 
			
			
		} else { 
			bzero(tmpbuf, sizeof(tmpbuf));
			snprintf(tmpbuf, sizeof(tmpbuf),
						"<TR><TD><FORM><INPUT TYPE=\"button\" VALUE=\"Delete\" onclick=\"parent.location='DELETE %s'\"></FORM></TD>\n", query_tmp->descriptor);

			strcat(outbuf, tmpbuf);
						
			bzero(tmpbuf, sizeof(tmpbuf));
			snprintf(tmpbuf, sizeof(tmpbuf),
						"<TD><FONT color=\"gray\" size=4>Query %d: %s</FONT><BR><BR></TD></TR>", i, tmp_desc);

			strcat(outbuf, tmpbuf);
				
		}

			i++;
	} //iterator

	pthread_mutex_unlock(&mutex);

}

/*
 * qrb_is_query_in_buffer()
 * 
 * Steps through the queries and checks for any matches.
 * 
 * Returns 0 if there is a match, -1 otherwise
 */
int qrb_is_query_in_buffer(char *query)
{
	GList *iterator;
	query_t *query_tmp;
	char *tmp_desc;	
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);

	for (iterator = query_report_buffer; iterator != NULL; 
								iterator = iterator->next) {
		query_tmp = (query_t *) iterator->data;
		tmp_desc = (char *) query_tmp->descriptor;
		
		if (strcasecmp(query, tmp_desc) == 0) {
			printf("qrb_is_query_in_buffer::Found matching query %s\n",
								query);
			return 0;
		}		
	}

	pthread_mutex_unlock(&mutex);
	return -1;
}

/*
 * qrb_create_query()
 *
 * Returns query is succesful, NULL otherwise
 */
query_t *qrb_create_query() 
{	
	query_t * query = (query_t *) malloc(sizeof(query_t));
	
	if (query == NULL) {
		fprintf(stderr, "qrb_create_query:: No memory for new query\n");
	} else {
		query_init(query);
	}

	return query;
}

/*
 * qrb_create_report()
 *
 * Returns new report if succesful (initialized), NULL otherwise
 */
report_t *qrb_create_report()
{	
	report_t *report = (report_t *) malloc(sizeof(report_t));
	
	if (report == NULL) {
		fprintf(stderr, "qrb_create_report:: No memory for new report\n");
	} else {
		report_init(report);
	}
	
	return report;
}

/*
 * qrb_add_report_to_query()
 * 
 * Adds reports to query
 */
void qrb_add_report_to_query(query_t *query, report_t *report)
{
		
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);

	
	/*
	 * FIXME: If report already exists, don't append
	 */
	if (qrb_is_report_in_query(query, report) == 0) {
		printf("Adding report %s to query %s\n", 
							report->local_url, 
							query->descriptor);
		qrb_remove_duplicate_reports(query, report);
		query->reports = g_list_append(query->reports, report);
	} else {
		printf("\nReport %s already in query %s\n",
						report->local_url, 
						query->descriptor);
	}	
		
	pthread_mutex_unlock(&mutex);
}
