/******************************************************************************
 *
 * 
 * smtp_server.c: This is the smtp server component of sevends, smtp clients
 * connect to the server, the servers save mail objects so can be transported
 * at a later stage.
 *
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.   
 * ***************************************************************************/
#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>           /* mkdir */
#include <sys/socket.h>         /* Network */
#include <netinet/in.h>
#include <strings.h>
#include <string.h>
#include <glib.h>

#include <libesmtp.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>


#include "main.h"
#include "email_objects_buffer.h"
#include "md5.h"
#include "global.h"

#define SMTP_LISTEN_PORT 9999
#define LISTEN_QUEUE 5
#define BUF_SIZE 1000
#define HEX_STRING "0123456789abcdef" /*to convert to hex */


#define MSG_220 "220 cs.columbia.edu ESMTP Sendmail (8.12.10/22/jtt/sed/ib42) is thrilled to serve you at Tue, 11 May 2004 17:36:08 -0400 (EDT). \r\n"

#define MSG_554 "554 noway \r\n"

#define MSG_250_1 "250-cs.columbia.edu greets localhost \r\n"
#define MSG_250_2 "250-8BITMIME \r\n"
#define MSG_250_3 "250-SIZE \r\n"
#define MSG_250_4 "250-DSN \r\n"
#define MSG_250_5 "250 HELP \r\n"
#define MSG_250_OK "250 OK \r\n"
#define MSG_354 "354 Start mail input; end with <CRLF>.<CRLF> \r\n"
#define MSG_221 "221 cs.columbia.edu Service closing transmission channel"

static void smtp_handle_request(int connfd);

/*
 * smtp_server_init() 
 *
 * This is the initialization call for the smtp thread
 */
int smtp_server_init() 
{
	pthread_t smtp_thread;
	int listenfd, connfd;
    socklen_t clilen;
    struct sockaddr_in client_addr, server_addr;

	printf("Smtp server init()\n");

    /* create server socket */
    if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        printf("SMTP: Error opening socket\n");
        return -1;
    }

    /* clear out struct */
    bzero(&server_addr, sizeof(server_addr));

	/* populate server address */
    server_addr.sin_family = AF_INET;
    server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    server_addr.sin_port = htons(SMTP_LISTEN_PORT); /*FIXME: Static now */

	
    if (bind
        (listenfd, (struct sockaddr *) &server_addr, sizeof(server_addr))
        < 0) {
        printf("SMTP_SERVER: Error binding socket\n");
        return -1;
    }

    if (listen(listenfd, LISTEN_QUEUE) < 0) {
        printf("Error listening socket\n");
        return -1;
    }

	for (;;) {
        clilen = sizeof(client_addr);
        connfd =
            accept(listenfd, (struct sockaddr *) &client_addr, &clilen);
	
		printf("connection!\n");
		/*
        if ((pthread_create(&smtp_thread,
                            NULL,
                            (void *(*)(void *)) smtp_handle_request,
                            (void *) &connfd)) != 0) {
            printf("Error -> ProxyServer: Error creating thread\n");
        }*/
		
		smtp_handle_request(connfd);
		close(connfd);
    }

    return 0;
}



/*
 * parse_from()
 *
 * Parse from field
 */
static int parse_from(char *from, email_object_t *object)
{
	char *beg;
	char *end;
	int from_len;
	char email[256];
	char *cp, *size_char;
	int size;
	
	if (!object) {
		fprintf(stderr, "parse_from: Object was empty\n");
		return -1;
	}

	/*
	 * Extract FROM
	 */
	beg = index(from, '<');
	end = rindex(from, '>');

	if ( (beg == NULL) || (end == NULL) ) {
		fprintf(stderr, "parse_from: Couldn't extract email\n");
		return -1;
	}

	bzero(email, sizeof(email));

	from_len = end - beg;
	
	strncpy(email, beg + 1, from_len - 1);

	email[from_len] = '\0';

	printf("FROM FIELD is %s\n", email);
	strncpy(object->mail_from, email, strlen(email));

	/*
	 * Extract BODY: FIXME: Hack for now
	 */
	if(strstr(from, "8BITMIME")) {
		printf("BODY is 8BITMIME \n");
		strncpy(object->body, "8BITMIME", strlen("8BITMIME"));
	}
	
	/*
	 * Extract SIZE
	 */
	cp = strstr(from, "SIZE=");

	if (!cp) {
		fprintf(stderr, "parse_from: Couldn't extract size\n");
		return -1;
	}
	
	size_char = cp + 5; /* 5 is SIZE= */
	size = atoi(size_char);

	printf("Size is %d\n", size);
	object->data_size = (size_t) size;

	return 0;	
}

/*
 * parse_to()
 *
 * Returns 0 on success -1 otherwise
 */
static int parse_to(char *to,  email_object_t *object) 
{
	char *beg;
	char *end;
	int to_len;
	char email[256];
	
	if (!object) {
		fprintf(stderr, "parse_from: Object was empty\n");
		return -1;
	}

	/*
	 * Extract TO
	 */
	beg = index(to, '<');
	end = rindex(to, '>');

	if ( (beg == NULL) || (end == NULL) ) {
		fprintf(stderr, "parse_to: Couldn't extract email\n");
		return -1;
	}

	bzero(email, sizeof(email));

	to_len = end - beg;
	
	strncpy(email, beg + 1, to_len - 1);

	email[to_len] = '\0';

	printf("TO FIELD is %s\n", email);
	strncpy(object->mail_to, email, strlen(email));
	
	return 0;

}
/*
 * MDPrint()
 * Converts digest to ASCII
 */
static void MDPrint(void *signature, char *str, const int str_len)
{

    unsigned char *sig_p;
    char *str_p, *max_p;
    unsigned int high, low;

    str_p = str;
    max_p = str + str_len;
    for (sig_p = (unsigned char *) signature;
         sig_p < (unsigned char *) signature + 16; sig_p++) {

        high = *sig_p / 16;
        low = *sig_p % 16;

        /* account for 2 chars */
        if (str_p + 1 >= max_p) {
            break;
        }

        *str_p++ = HEX_STRING[high];
        *str_p++ = HEX_STRING[low];
    }
    if (str_p < max_p) {
        *str_p++ = '\0';
    }
}

/*
 * MDString()
 * Creates a hexadecimal string representation
 */
static void MDString(char *string, char *md5)
{
    MD5_CTX context;
    unsigned char digest[16];
    unsigned int len = strlen(string);

    MD5Init(&context);
    MD5Update(&context, string, len);
    MD5Final(digest, &context);

    MDPrint(&(digest[0]), md5, 16);
}

/*
 * smtp_handle_request()
 *
 * Handles the requests coming in for smtp.
 *
 * This should be quite straight forward as the server will
 * most likely accept all request and make transformations in
 * the from field as most smtp relays don't allow for address 
 * fwding from addresses that are not in their domain.
 *
 * 		Sample request exchange for a successful transaction
 *
 * 		
 * 	  S: 220 foo.com Simple Mail Transfer Service Ready
      C: EHLO bar.com
      S: 250-foo.com greets bar.com
      S: 250-8BITMIME
      S: 250-SIZE
      S: 250-DSN
      S: 250 HELP
      C: MAIL FROM:<Smith@bar.com>
      S: 250 OK
      C: RCPT TO:<Jones@foo.com>
      S: 250 OK
      C: RCPT TO:<Green@foo.com>
      S: 550 No such user here
      C: RCPT TO:<Brown@foo.com>

      S: 250 OK
      C: DATA
      S: 354 Start mail input; end with <CRLF>.<CRLF>
      C: Blah blah blah...
      C: ...etc. etc. etc.
      C: .
      S: 250 OK
      C: QUIT
      S: 221 foo.com Service closing transmission channel


	  =================

	  Mail From: MAIL FROM:<stelios@cs.columbia.edu> BODY=8BITMIME SIZE=370

	RCPT TO: RCPT TO:<douskos@hotmail.com>

	Data? DATA

	Data is From: Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
	To: douskos@hotmail.com
	Subject: hello
	Date: Tue, 11 May 2004 22:29:21 -0400
	User-Agent: KMail/1.6.2
	MIME-Version: 1.0
	Content-Disposition: inline
	Content-Type: text/plain;
  	charset="us-ascii"
	Content-Transfer-Encoding: 7bit
	Message-Id: <200405112229.21435.stelios@cs.columbia.edu>

 */
static void smtp_handle_request(int connfd)
{
	int fd = (int) connfd;
	int c = 0;
	char buf[BUF_SIZE + 1];
	char *data_buf;
	email_object_t *object;

	/*
	 * First create the object that will hold the email
	 */
	object = (email_object_t *) email_create_object();

	if (!object) {
		fprintf(stderr,
				"Error->smtp_handle_request : couldn't create object\n");
	}

	/*
	 * 220
	 */
	if ((c = send(fd, MSG_220, strlen(MSG_220), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	/*
	 * EHLO
	 */
	if ((c = recv(fd, buf, BUF_SIZE, 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }
	//printf("I received: %s", buf);

	/*
	 * 250's 
	 */
	if ((c = send(fd, MSG_250_1, strlen(MSG_250_1), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }
	
	if ((c = send(fd, MSG_250_2, strlen(MSG_250_2), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	if ((c = send(fd, MSG_250_3, strlen(MSG_250_3), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	if ((c = send(fd, MSG_250_4, strlen(MSG_250_4), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	if ((c = send(fd, MSG_250_5, strlen(MSG_250_5), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	/*
	 * Mail From
	 */
	bzero(buf, BUF_SIZE);
	if ((c = recv(fd, buf, BUF_SIZE, 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }
	//printf("Mail From: %s\n", buf);
	parse_from(buf, object);

	if ((c = send(fd, MSG_250_OK, strlen(MSG_250_OK), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	/*
	 * RCPT TO:
	 */
	bzero(buf, BUF_SIZE);
	if ((c = recv(fd, buf, BUF_SIZE, 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }
	//printf("RCPT TO: %s\n", buf);

	parse_to(buf, object);
	
	if ((c = send(fd, MSG_250_OK, strlen(MSG_250_OK), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	/*
	 * DATA
	 */
	bzero(buf, BUF_SIZE);
	if ((c = recv(fd, buf, BUF_SIZE, 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }
	
	if ((c = send(fd, MSG_354, strlen(MSG_354), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }
	//printf("Data? %s\n", buf);
	//if (strcmp(buf, "DATA"

	
	data_buf = malloc(object->data_size);
	if (!data_buf) {
		fprintf(stderr,
					"Error->smtp_handle_request : Couldn't allocate "
					" enough memory for data_buf\n");
		return;
	}
	
	bzero(data_buf, object->data_size);
	if ((c = recv(fd, data_buf, object->data_size, 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }
	//printf("Data is %s\n Length is %d\n", buf, strlen(buf));

	/* Add data to object */
	object->data = malloc(object->data_size);
	if (!object->data) {
		fprintf(stderr, 
			"Error->smtp_handle_request: Couldn't Allocate mem for object\n");
		return;
	}
	memcpy(object->data, data_buf, object->data_size);

	free(data_buf);

	if ((c = send(fd, MSG_250_OK, strlen(MSG_250_OK), 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	/*
	 * QUIT
	 */
	bzero(buf, BUF_SIZE);
	if ((c = recv(fd, buf, BUF_SIZE, 0) < 0)) {
        fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
        return;
    }

	if (strncmp(buf, "QUIT", 4) == 0) {
		//printf("QUIT!\n");
		if ((c = send(fd, MSG_221, strlen(MSG_221), 0) < 0)) {
        	fprintf(stderr, 
					"Error->smtp_handle_request : Couldn't read socket \n");
    	    return;
	    }

	}

	/*
	 * Create MD5
	 */
	MDString(object->data, object->md5);

	/*
	 * Now add object to linked list
	 */
	email_append_object(object);

}



