/******************************************************************************
 *
 * swish_manager.c
 * 
 * A search engine manager. Responsible for creating and maintaining the index
 * of the cached web pages.
 * 
 * http://swish-e.org/
 * 
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/

#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef ARM
#include "libs/swish-e.h"
#else 
#include <swish-e.h>
#endif

#define DEBUG(x) printf(x)
#define DISPLAY_COUNT 10  /* max to display */
#define INDEX "swish-e -v 0 -c configuration/swish.conf"
#define UPDATE_INDEX "swish-e -v 0 -c configuration/swish.conf"

int swish_index_period = 10; /* seconds */

/*
 * get_swish_index_period()
 *
 * Returns swish index period in seconds
 */
int get_swish_index_period() 
{	
	return swish_index_period;
}


/*
 * set_swish_index_period()
 *
 * Set indexing period in seconds
 */
void set_swish_index_period(int new_time)
{
	if (new_time > 0)
		swish_index_period = new_time;
}

/*
 * update_index()
 * 
 * Responsible for updating the index periodically
 */
void swish_update_index()
{
	/*Update index then sleep*/
		
	for (;;) {
		DEBUG("swish_manager: Updating index\n");
		system(UPDATE_INDEX);
		sleep(swish_index_period);
	}
	
	
}

/*
 * swish_manager_init
 *
 * Swish initialization
 */
void swish_manager_init(char *filename)
{
	pthread_t swish_thread;

	/* 
	 * First create index if it's not there then loop for given time 
	 * updating the index
	 */
	
	system(INDEX);
	
	/*Launch update index thread */
	if ((pthread_create(&swish_thread,
                            NULL,
                            (void *(*)(void *)) swish_update_index,
                            NULL)) != 0) {
            printf("Error -> swish_manager_init: Error creating thread\n");
	}
}

