/******************************************************************************
 *
 * XMLPARSER
 *
 * 
 * ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>

//#include <libxml/parser.h>
//#include <libxml/xmlmemory.h>
#include "main.h"
#include "query_report_buffer.h"
#include "xmlparser.h"
#include "caching.h"
#include "parser.h"
#include "referer_hash.h"

#define DEBUG(x) printf(x)
#define XML_ROOT "SEVENDS_QUERY"

static void parse_report(xmlDocPtr doc, xmlNodePtr cur, query_t *query); 
static query_t *parse_query_name(xmlDocPtr doc, 
						xmlNodePtr cur);


/*
 * parse_multicast_xml()
 * Responsible for parsing the query xml message
 *
 * Returns 0 on success, -1 on error.
 */
int parse_multicast_xml(char *mesg, char* request)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	//xmlNodePtr query_cur;
	xmlChar *key;

	DEBUG("parse_multicast_xml\n");

	if ((strlen(mesg) == 0) || mesg == NULL) {
		fprintf(stderr,
					"parse_multicast_xml:: mesg is empty.\n");
		return -1;

	}
	
	doc = xmlParseMemory(mesg, strlen(mesg));

	if (doc == NULL ) {
		fprintf(stderr,
			"parse_multicast_xml:: %s not parsed successfully.\n", mesg);
		return -1;
	}

	cur = xmlDocGetRootElement(doc);
	
	if (cur == NULL) {
		fprintf(stderr,"parse_multicast_xml:: Empty document\n");
		xmlFreeDoc(doc);
		return -1;
	}
	
	/*Get root name */
	if (xmlStrcmp(cur->name, (const xmlChar *) XML_ROOT)) {
		fprintf(stderr,
				"parse_multicast_xml:: document of the wrong type");
		xmlFreeDoc(doc);
		return -1;
	}

	/* Get children of SEVENDS QUERY */
	cur = cur->xmlChildrenNode;

	while (cur != NULL) {
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Query"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		    //request = malloc(strlen(key)); //
			strncpy(request, key, strlen(key));
			xmlFree(key);

		}
		
	   	/* Iterate */	
		cur = cur->next;
	}
	
	xmlFreeDoc(doc);
	return 0; /*success*/


}

/*
 * parse_unicast_xml()
 * This method parses the xml report mesg that comes in on the unicast_receiver
 * returns -1 on failure, 0 on success
 */
int parse_unicast_xml(char *mesg)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	xmlNodePtr query_cur;
	//report_t *report;
	query_t *query = NULL;

	DEBUG("parse_unicast_xml::\n");
	
	doc = xmlParseMemory(mesg, strlen(mesg));
	
	if ((strlen(mesg) == 0) || mesg == NULL) {
		fprintf(stderr,
					"parse_unicast_xml:: %s not parsed successfully.\n", mesg);
		return -1;

	}

	if (doc == NULL ) {
		fprintf(stderr,
					"parse_unicast_xml:: %s not parsed successfully.\n", mesg);
		return -1;
	}

	cur = xmlDocGetRootElement(doc);
	
	if (cur == NULL) {
		fprintf(stderr,"parse_unicast_xml:: Empty document\n");
		xmlFreeDoc(doc);
		return -1;
	}
	
	/*Get root name */
	if (xmlStrcmp(cur->name, (const xmlChar *) XML_ROOT)) {
		fprintf(stderr,
				"parse_unicast_xml:: document of the wrong type, root node != story");
		xmlFreeDoc(doc);
		return -1;
	}

		
	/* Get children of SEVENDS QUERY */
	cur = cur->xmlChildrenNode;
	
	while (cur != NULL) {
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Query"))) {
			
			/* Get child first before looking for QueryName */
			query_cur = cur->children;
			if ((!xmlStrcmp(query_cur->name, (const xmlChar *)"QueryName"))) {
					query = parse_query_name(doc, query_cur);
					if (query == NULL) continue;
			}
			query_cur = query_cur->next;
			while (query_cur != NULL) {
				if ((!xmlStrcmp(query_cur->name, (const xmlChar *)"Report"))) {
					parse_report(doc, query_cur, query);
					//add_report_to_query(query, report);
				}
				query_cur = query_cur->next;
			}
		}
		
	   	/* Iterate */	
		cur = cur->next;
	}
	
	xmlFreeDoc(doc);
	return 0; /*success*/

}

/*
 * create_query_xml()
 *
 * Returns 0 on success, -1 on failure
 */
int create_query_xml(char *keyword, char *xml)
{
	xmlDocPtr doc;
	//xmlNodePtr cur;
	xmlNodePtr query_cur;
	xmlNodePtr root;
	xmlChar *mem;
	int size;

	printf("create_query\n");
	
	if ((strlen(keyword) == 0) ||  keyword == NULL) {
		fprintf(stderr,
					"create_query_xml:: %s not parsed successfully.\n", keyword);
		return -1;

	}

	doc = xmlNewDoc((const xmlChar *) "1.0");
	
	if (doc == NULL ) {
		fprintf(stderr,
				"create_query_xml:: Document not created successfully.\n");
		return -1;
	}

	root = xmlNewNode(NULL, (const xmlChar *) XML_ROOT);
	
	xmlDocSetRootElement(doc, root);
	
	if (root == NULL) {
			fprintf(stderr,
						"create_query_xml:: Root node not created\n");
		return -1;

	}
	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Query", (const xmlChar *)keyword);

	//xmlSaveFile("test.xml", doc);
	xmlDocDumpMemory(doc, &mem, &size);
	
	/*Make sure you free xml after use */
	//xml = malloc(size + 1);
	//	if (sizeof(xml) < size)
	//		realloc(xml, size);
	
	strncpy(xml, mem, size+1); 
	printf("Mem contents are \n%s\n", xml);
	xmlFreeDoc(doc);

	return 0;
}

/*
 * parse_report()
 * This method parses the internals of the report. To parse new fields just
 * copy the if statement below substituting the data structure.
 */
static void parse_report(xmlDocPtr doc, xmlNodePtr cur, query_t *query) 
{
	xmlChar *key;
	report_t * report;
	int rank, size;
	time_t time;
	
	cur = cur->xmlChildrenNode;
	
	DEBUG("parse_report::\n");
	
	report = qrb_create_report();

	while (cur != NULL) {
	    		
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Title"))) {
	    	key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			
		    printf("Title: %s\n", key);
			if (key != NULL)
				strncpy(report->title, key, sizeof(report->title));
		 	else 
				strcpy(report->title,"No title");	
	
			xmlFree(key);
 	    } 
		
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Orig-URL"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		    printf("Orig-URL: %s\n", key);
			strncpy(report->orig_url, key, sizeof(report->orig_url)); 
			xmlFree(key);
		}

		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Local-URL"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		    printf("Local-URL: %s\n", key);
			if (key != NULL) {
				strncpy(report->local_url, key, sizeof(report->local_url)); 
				referer_create_referer(report->local_url);
			} else { 
				strcpy(report->local_url, "No Local URL");
			}
		    xmlFree(key);
		}

		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Date"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			time = atoi(key);
			printf("Date: %s\n", ctime(&time));
			strncpy(report->date, ctime(&time), sizeof(report->date));
			report->age = time;
			
		    xmlFree(key);
		}

		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Summary"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			if (strlen(key) > 1) printf("Summary: %s\n", key);
			strncpy(report->summary, key, sizeof(report->summary));			
		    xmlFree(key);
		}

		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Rank"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			if (strlen(key) > 1) { 
				printf("Rank: %s\n", key);
				rank = atoi(key);
				report->rank = rank;
			}
		    xmlFree(key);
		}
	
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Size"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			if (strlen(key) > 1) { 
				printf("Size: %s\n", key);
				size = atoi(key);
				report->size = size;
			}
		    xmlFree(key);
		}

		

		cur = cur->next; /* Get next report element */
	}
	
	qrb_add_report_to_query(query, report);	
	
	printf("\n"); /* Separate reports */			

    return ;

}


/*
 * parse_query_name()
 * Parses the query name
 */
static query_t *parse_query_name(xmlDocPtr doc, 
						xmlNodePtr cur)
{	
	xmlChar *key;
	GSList *iterator = NULL; /*iterator temp */
	query_t *query_tmp;
	int is_new = 1;
	char *desc;

	key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	printf("QueryName: %s\n", key);
	printf("----------------------------------------\n");
	/*FIXME:
	 * If the query descriptor does not exist, add query. 
	 * Otherwise add reports to this query.
	 */

	for (iterator = qrb_get_buffer(); 
					iterator != NULL; iterator =iterator->next) {
		query_tmp = (query_t *) iterator->data;
		desc = query_tmp->descriptor;
		printf("Current Descriptor : %s\n Key: %s", desc, key);
		/*
		 * Here we assume that there's only query with the same
		 * descriptor. It should be correct if we check when creating
		 * new queries
		 */
		if (strcmp(desc, key) == 0) {
			/*query already exists, add reports to this query */
			query_tmp = (query_t *)iterator->data;
		   	is_new = 0;
			break;	
		} 
	}
	
	if (is_new) {
		/*
	 	 * I allocate the mem here for the queries but I am responsible
	 	 * for cleaning up when a query is deleted... 
	 	 */
		query_tmp = qrb_create_query();
		if (!query_tmp) {
			fprintf(stderr, "parse_doc:: Error allocating mem for query\n");
			xmlFree(key);
			return NULL;
		}
		strncpy(query_tmp->descriptor, key, sizeof(query_tmp->descriptor));
		qrb_append_queries(query_tmp);
			
	}


	/* Always free key */
	xmlFree(key);
    return query_tmp;

}
/*
 * Test method for parsing an xml file
 */
int test_parse_doc(char *filename, struct data_structures *data)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	xmlNodePtr query_cur;
	query_t *query;

	printf("test_parse_doc...\n");

	doc = xmlParseFile(filename);

	if (doc == NULL ) {
		fprintf(stderr,
						"parse_doc:: %s not parsed successfully.\n", filename);
		return -1;
	}

	cur = xmlDocGetRootElement(doc);
	
	if (cur == NULL) {
		fprintf(stderr,"parse_doc:: Empty document\n");
		xmlFreeDoc(doc);
		return -1;
	}
	
	/*Get root name */
	if (xmlStrcmp(cur->name, (const xmlChar *) XML_ROOT)) {
		fprintf(stderr,
				"parse_doc:: document of the wrong type, root node != story");
		xmlFreeDoc(doc);
		return -1;
	}

	/* Get children of SEVENDS QUERY */
	cur = cur->xmlChildrenNode;
	
	while (cur != NULL) {
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"Query"))) {
			
			/* Get child first before looking for QueryName */
			query_cur = cur->children;
			if ((!xmlStrcmp(query_cur->name, (const xmlChar *)"QueryName"))) {
					query = parse_query_name(doc, query_cur);
					if (query == NULL) return -1;
			}
			query_cur = query_cur->next;
			while (query_cur != NULL) {
				if ((!xmlStrcmp(query_cur->name, (const xmlChar *)"Report"))) {
					parse_report(doc, query_cur, query);
				}
				query_cur = query_cur->next;
			}
		}
		
	   	/* Iterate */	
		cur = cur->next;
	}
	
	xmlFreeDoc(doc);
	return 0; /*success*/

	
}

/*
 * create_response_xml()
 *  
 * Create an xml response based on the url that exists on our cache
 * 
 * return 0 on success -1 otherwise.
 */
int create_response_xml(char *path, char*xml_response)
{
	/*FIXME: Fill me in */
	DEBUG("create_response_xml\n");

	return 0;
}


/*
 * Test program
 */
void parseStory (xmlDocPtr doc, xmlNodePtr cur) {

	xmlChar *key;
	cur = cur->xmlChildrenNode;
	//cur = cur->next;
	while (cur != NULL) {
	    if ((!xmlStrcmp(cur->name, (const xmlChar *)"QueryName"))) {
	    	key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		    printf("QueryName: %s\n", key);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"author"))) {
			key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		    printf("author: %s\n", key);
		    xmlFree(key);
		}
		
		cur = cur->next;
	}
    return;
}


/***********************GET_REPORT_XML******************************/


/*
 * create_xml_query_name()
 *
 * Returns 0 on success, -1 on failure
 */
int create_xml_query_name(xmlNodePtr root, xmlNodePtr query_cur, char *keyword)
{
	DEBUG("create_xml_query_name\n");
	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "QueryName", (const xmlChar *)keyword);

	if (query_cur == NULL)
		return -1;
	else 
		return 0;

}

/*
 * create_xml_query()
 * 
 * Returns 0 on success, -1 on failure
 */
xmlNodePtr create_xml_query(xmlNodePtr root, xmlNodePtr query_cur, char *keyword)
{
	DEBUG("create_xml_query\n");
	return xmlNewChild(root, NULL, (const xmlChar *) "Query", NULL);
}


/*
 * create_xml_local_url()
 * 
 * Returns 0 on success, -1 on failure
 */
int create_xml_local_url(xmlNodePtr root, xmlNodePtr query_cur, char *local_url)
{	
	char *path;

	DEBUG("create_xml_local_url\n");

	path = malloc(strlen(local_url) + 7 /* 7 for http:// */
							+ 16 /* IP address */
							+ 7);	/* :PORT/ */
	bzero(path, sizeof(path));

	local_url = cache_strip_path(local_url);
	path = add_ip_and_port(local_url, path);

	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Local-URL", (const xmlChar *)path);

	free(path);
	if (query_cur == NULL)
		return -1;
	else 
		return 0;
}

/*
 * create_xml_original_url()
 * 
 * Returns 0 on success, -1 on failure
 */
int create_xml_original_url(xmlNodePtr root, xmlNodePtr query_cur, char *orig_url)
{
	DEBUG("create_xml_original_url\n");

	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Orig-URL", (const xmlChar *)orig_url);

	if (query_cur == NULL)
		return -1;
	else 
		return 0;
}

/*
 * create_xml_title()
 * Returns 0 on success, -1 on failure
 */
int create_xml_title(xmlNodePtr root, xmlNodePtr query_cur, char *title)
{
	DEBUG("create_xml_title\n");

	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Title", (const xmlChar *)title);

	if (query_cur == NULL)
		return -1;
	else 
		return 0;

}

/*
 * create_xml_date()
 * Returns 0 on success, -1 on failure
 */
int create_xml_date(xmlNodePtr root, xmlNodePtr query_cur, char *date)
{	
	DEBUG("create_xml_title\n");

	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Date", (const xmlChar *)date);

	if (query_cur == NULL)
		return -1;
	else 
		return 0;
}

/*
 * create_xml_rank()
 *
 * Adds Rank to query
 *
 * Returns 0 on success, -1 on failure
 */
int create_xml_rank(xmlNodePtr root, xmlNodePtr query_cur, char *rank)
{
	DEBUG("create_xml_rank\n");

	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Rank", (const xmlChar *)rank);

	if (query_cur == NULL)
		return -1;
	else 
		return 0;
}
/*
 * create_xml_size()
 *
 * Adds size to query
 *
 * Returns 0 on success, -1 on failure
 */

int create_xml_size(xmlNodePtr root, xmlNodePtr query_cur, char *size)
{
	DEBUG("create_xml_size\n");

	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Size", (const xmlChar *)size);

	if (query_cur == NULL)
		return -1;
	else 
		return 0;
}



/*
 * create_xml_report()
 * 
 * Returns 0 on success, -1 on failure
 */
int create_xml_report(xmlNodePtr root, xmlNodePtr query_cur, char *result)
{
	DEBUG("create_xml_report\n");
	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Report", NULL);

	/*
	 * The result would need to be parsed for other fields.
	 */
	create_xml_local_url(query_cur, query_cur, result);

	return 0;
}


/*
 * create_xml_root_doc()
 * 
 * Creates the root element for the xml document
 * 
 * Returns 0 on success, -1 on failure
 */
int create_xml_root_doc(xmlDocPtr doc, xmlNodePtr root)
{
	DEBUG("create_xml_root_doc\n");

	/*
	 * Create XML doc
	 */
	doc = xmlNewDoc((const xmlChar *) "1.0");
	
	if (doc == NULL ) {
		fprintf(stderr,
				"get_report:: Document not created successfully.\n");
		return -1;
	}

	root = xmlNewNode(NULL, (const xmlChar *) XML_ROOT);
	
	xmlDocSetRootElement(doc, root);
	
	if (root == NULL) {
			fprintf(stderr,
						"get_report:: Root node not created\n");
		return -1;

	}

	return 0;
}

